#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 17 08:55:47 2017

@author: marta
"""

from astropy import log
from stingray.utils import excess_variance
from .io import load_lcurve
from .io import save_as_qdp
from .base import hen_root


def fvar(lc):
    return excess_variance(lc, normalization="fvar")


def excvar_none(lc):
    return excess_variance(lc, normalization="none")


def excvar_norm(lc):
    return excess_variance(lc, normalization="norm_xs")


def main(args=None):
    import argparse
    from .base import _add_default_args, check_negative_numbers_in_args

    description = "Calculate excess variance in light curve chunks"
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of files", nargs="+")
    parser.add_argument(
        "-c",
        "--chunk-length",
        type=float,
        default=20,
        help="Length in seconds of the light curve chunks",
    )
    parser.add_argument(
        "--fraction-step",
        type=float,
        default=0.5,
        help="If the step is not a full chunk_length but less,"
        "this indicates the ratio between step step and"
        " `chunk_length`",
    )
    parser.add_argument(
        "--norm",
        type=str,
        default="excvar",
        help="Choose between fvar, excvar and norm_excvar "
        "normalization, referring to Fvar, excess "
        "variance, and normalized excess variance "
        "respectively (see"
        " Vaughan et al. 2003 for details).",
    )
    _add_default_args(parser, ["loglevel", "debug"])

    args = check_negative_numbers_in_args(args)
    args = parser.parse_args(args)

    if args.debug:
        args.loglevel = "DEBUG"

    log.setLevel(args.loglevel)
    with log.log_to_file("HENexcvar.log"):
        filelist = []
        for fname in args.files:
            lcurve = load_lcurve(fname)
            if args.norm == "fvar":
                start, stop, res = lcurve.analyze_lc_chunks(
                    args.chunk_length, fvar, args.fraction_step
                )
            elif args.norm == "excvar":
                start, stop, res = lcurve.analyze_lc_chunks(
                    args.chunk_length, excvar_none, args.fraction_step
                )
            elif args.norm == "norm_excvar":
                start, stop, res = lcurve.analyze_lc_chunks(
                    args.chunk_length, excvar_norm, args.fraction_step
                )
            else:
                raise ValueError(
                    "Normalization must be fvar, norm_excvar " "or excvar"
                )
            var, var_err = res
            out = hen_root(fname) + "_" + args.norm + ".qdp"
            save_as_qdp(
                [(start + stop) / 2, var],
                [(stop - start) / 2, var_err],
                filename=out,
            )
            filelist.append(out)

    return filelist



# -*- coding: utf-8 -*-

from astropy import log
from .io import load_pds
from .io import save_as_qdp
from .base import hen_root


def main(args=None):
    import argparse
    from .base import _add_default_args

    description = (
        "Read timelags from cross spectrum results and save them"
        " to a qdp file"
    )
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of files", nargs="+")

    _add_default_args(parser, ["loglevel", "debug"])

    args = parser.parse_args(args)

    if args.debug:
        args.loglevel = "DEBUG"

    log.setLevel(args.loglevel)
    with log.log_to_file("HENlags.log"):
        filelist = []
        for fname in args.files:
            cross = load_pds(fname)

            lag, lag_err = cross.time_lag()
            out = hen_root(fname) + "_lags.qdp"
            save_as_qdp([cross.freq, lag], [None, lag_err], filename=out)
            filelist.append(out)

    return filelist



"""Interactive phaseogram."""

import copy
import argparse
import warnings
from abc import abstractmethod

from scipy.interpolate import interp1d
import numpy as np
from astropy import log
from astropy.logger import AstropyUserWarning
from astropy.stats import poisson_conf_interval
from stingray.pulse.search import phaseogram
from stingray.utils import assign_value_if_none
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button
from matplotlib.gridspec import GridSpec
from scipy.ndimage import gaussian_filter1d

from .fold import filter_energy
from .io import load_events, load_folding
from .fold import get_TOAs_from_events
from .base import hen_root, deorbit_events
from .efsearch import h_test


DEFAULT_COLORMAP = "viridis"


def get_H_label(phas, prof):
    good = phas < 1
    nmax = max(len(prof[good]) // 3, 1)
    prof_h, prof_m = h_test(prof[good], nmax=nmax)
    # z2_detlev = z2_n_detection_level(n=2)
    h_label = f"H test = {prof_h:.1f} (M={prof_m})"
    return h_label


def sec_to_mjd(time, mjdref):
    return time / 86400 + mjdref


class SliderOnSteroids(Slider):
    def __init__(self, *args, **kwargs):
        self.hardvalmin = None
        self.hardvalmax = None
        if "hardvalmin" in kwargs:
            self.hardvalmin = kwargs["hardvalmin"]
            kwargs.pop("hardvalmin")
        if "hardvalmax" in kwargs:
            self.hardvalmax = kwargs["hardvalmax"]
            kwargs.pop("hardvalmax")
        Slider.__init__(self, *args, **kwargs)


def normalized_phaseogram(norm, *args, **kwargs):
    phas, phases, times, additional_info = phaseogram(*args, **kwargs)
    if norm is None:
        pass
    elif norm == "to1":
        minarr = np.min(phas, axis=0)
        maxarr = np.max(phas, axis=0)
        for i in range(phas.shape[0]):
            phas[i][:] -= minarr
            phas[i][:] /= maxarr - minarr
    elif norm == "mediansub":
        medarr = np.median(phas, axis=0)
        for i in range(phas.shape[0]):
            phas[i][:] -= medarr
    elif norm == "mediannorm":
        medarr = np.median(phas, axis=0)
        for i in range(phas.shape[0]):
            phas[i][:] -= medarr
        phas[i][:] /= medarr
    elif norm == "meansub":
        medarr = np.mean(phas, axis=0)
        for i in range(phas.shape[0]):
            phas[i][:] -= medarr
    elif norm == "meannorm":
        medarr = np.mean(phas, axis=0)
        for i in range(phas.shape[0]):
            phas[i][:] -= medarr
        phas[i][:] /= medarr
    else:
        warnings.warn(
            "Profile normalization " "{} not known. Using default".format(norm)
        )
    return phas, phases, times, additional_info


class BasePhaseogram(object):
    def __init__(
        self,
        ev_times,
        freq,
        nph=128,
        nt=128,
        test=False,
        fdot=0,
        fddot=0,
        mjdref=None,
        pepoch=None,
        gti=None,
        label="phaseogram",
        norm=None,
        position=None,
        object=None,
        plot_only=False,
        time_corr=None,
        **kwargs,
    ):
        """Init BasePhaseogram class.

        Parameters
        ----------
        ev_times : array-like
            Event times
        freq : float
            Frequency of pulsation

        Other parameters
        ----------------
        nph : int
            Number of phase bins in the profile
        nt : int
            Number of time bins in the profile
        pepoch : float, default None
            Epoch of timing solution, in the same units as ev_times
        mjdref : float, default None
            Reference MJD
        fdot : float
            First frequency derivative
        fddot : float
            Second frequency derivative
        label : str
            Label for windows
        norm : str
            Normalization
        position : `astropy.Skycoord` object
            Position of the pulsar
        object : str
            Name of the pulsar
        **kwargs : keyword args
            additional arguments to pass to `self._construct_widgets`
        """

        self.fdot = fdot
        self.fddot = fddot
        self.nt = nt
        self.nph = nph
        if mjdref is None:
            warnings.warn("MJDREF not set. All MJD values will be incorrect.")
        self.mjdref = assign_value_if_none(mjdref, 0.0)
        self.ev_times = np.asarray(ev_times)
        self.gti = gti
        if gti is None:
            self.gti = np.array([[ev_times[0], ev_times[-1]]])
        self.label = label
        self.test = test

        self.pepoch = assign_value_if_none(pepoch, ev_times[0])

        self.time_corr = assign_value_if_none(
            time_corr, np.zeros_like(ev_times)
        )
        self.freq = freq
        self.norm = norm
        self.position = position
        self.object = object
        self.timing_model_string = ""

        self.time_corr_fun = interp1d(
            self.ev_times,
            self.time_corr,
            bounds_error=False,
            fill_value="extrapolate",
        )
        self.time_corr_mjd_fun = interp1d(
            self.ev_times / 86400 + self.mjdref,
            self.time_corr / 86400,
            bounds_error=False,
            fill_value="extrapolate",
        )

        self.fig = plt.figure(label, figsize=(6, 8))
        gs = GridSpec(3, 1, height_ratios=[2, 3, 0.2])
        plt.subplots_adjust(left=0.1, bottom=0.30, top=0.95, right=0.95)
        ax = plt.subplot(gs[1])
        self.profax = plt.subplot(gs[0], sharex=ax)
        self.profax.set_xticks([])
        colorbax = plt.subplot(gs[2])

        corrected_times = self.ev_times - self._delay_fun(self.ev_times)
        (
            self.unnorm_phaseogr,
            phases,
            times,
            additional_info,
        ) = normalized_phaseogram(
            None,
            corrected_times,
            freq,
            return_plot=True,
            nph=nph,
            nt=nt,
            fdot=fdot,
            fddot=fddot,
            plot=False,
            pepoch=self.pepoch,
        )

        self.phaseogr, phases, times, additional_info = normalized_phaseogram(
            self.norm,
            corrected_times,
            freq,
            return_plot=True,
            nph=nph,
            nt=nt,
            fdot=fdot,
            fddot=fddot,
            plot=False,
            pepoch=self.pepoch,
        )

        self.phases, self.times = phases, times
        vmin = None
        if self.norm is not None and (
            "sub" in self.norm or "to1" in self.norm
        ):
            vmin = 0
        self.pcolor = ax.pcolormesh(
            phases, times, self.phaseogr.T, cmap=DEFAULT_COLORMAP, vmin=vmin
        )
        self.colorbar = plt.colorbar(
            self.pcolor, cax=colorbax, orientation="horizontal"
        )
        ax.set_xlabel("Phase")

        def s2d(x):
            return (x - self.pepoch) / 86400

        def d2s(x):
            return (x - self.mjdref) * 86400

        secax = ax.secondary_yaxis("left", functions=(s2d, d2s))
        secax.set_ylabel(f"d from MJD {self.pepoch / 86400 + self.mjdref}")

        plt.setp(ax.get_yticklabels(), visible=False)
        ax.set_yticks([])
        # plt.setp(ax.get_yticks(), visible=False)

        # ax.set_ylabel('MJD')
        # plt.colorbar()
        self.lines = []
        self.line_phases = np.arange(-2, 3, 0.5)
        for ph0 in self.line_phases:
            (newline,) = ax.plot(
                np.zeros_like(times) + ph0, times, zorder=10, lw=2, color="w"
            )
            self.lines.append(newline)

        ax.set_xlim([0, 2])

        axcolor = "#ff8888"
        self.slider_axes = []

        def newax_fn(*args, **kwargs):
            ax = self.fig.add_axes(*args, facecolor=axcolor)
            return ax

        self.slider_axes.append(
            newax_fn([0.15, 0.1, 0.75, 0.03], facecolor=axcolor)
        )
        self.slider_axes.append(
            newax_fn([0.15, 0.15, 0.75, 0.03], facecolor=axcolor)
        )
        self.slider_axes.append(
            newax_fn([0.15, 0.2, 0.75, 0.03], facecolor=axcolor)
        )

        self._construct_widgets(**kwargs)

        self.closeax = self.fig.add_axes([0.15, 0.020, 0.15, 0.04])
        self.button_close = Button(
            self.closeax, "Quit", color=axcolor, hovercolor="0.8"
        )

        self.recalcax = self.fig.add_axes([0.3, 0.020, 0.15, 0.04])
        self.button_recalc = Button(
            self.recalcax, "Recalculate", color=axcolor, hovercolor="0.975"
        )

        self.resetax = self.fig.add_axes([0.45, 0.020, 0.15, 0.04])
        self.button_reset = Button(
            self.resetax, "Reset", color=axcolor, hovercolor="0.975"
        )

        self.zoominax = self.fig.add_axes([0.6, 0.020, 0.1, 0.04])
        self.button_zoomin = Button(
            self.zoominax, "+Zoom", color=axcolor, hovercolor="0.975"
        )

        self.zoomoutax = self.fig.add_axes([0.7, 0.020, 0.1, 0.04])
        self.button_zoomout = Button(
            self.zoomoutax, "-Zoom", color=axcolor, hovercolor="0.975"
        )

        self.toaax = self.fig.add_axes([0.8, 0.020, 0.1, 0.04])
        self.button_toa = Button(
            self.toaax, "TOA", color=axcolor, hovercolor="0.975"
        )

        self.button_reset.on_clicked(self.reset)
        self.button_zoomin.on_clicked(self.zoom_in)
        self.button_zoomout.on_clicked(self.zoom_out)
        self.button_toa.on_clicked(self.toa)
        self.button_recalc.on_clicked(self.recalculate)
        self.button_close.on_clicked(self.quit)
        # self.profax = plt.axes([0.25, 0.75, 0.5, 0.2])

        prof = np.sum(np.nan_to_num(self.unnorm_phaseogr), axis=1)
        nbin = len(prof)
        phas = np.linspace(0, 2, nbin + 1)[:-1]
        (self.profile_fixed,) = self.profax.plot(
            phas, prof, drawstyle="steps-post", color="grey"
        )
        (self.profile,) = self.profax.plot(
            phas, prof, drawstyle="steps-post", color="k"
        )
        mean = np.mean(prof)
        low, high = poisson_conf_interval(
            mean, interval="frequentist-confidence", sigma=2
        )
        self.profax.fill_between(phas, low, high, alpha=0.5)
        z2_label = get_H_label(phas, prof)
        self.proftext = self.profax.text(
            0.1, 0.8, z2_label, transform=self.profax.transAxes
        )
        if not test and not plot_only:
            plt.show()
        if plot_only:
            plt.savefig(self.label + "_{:.10f}Hz.png".format(self.freq))

    @abstractmethod
    def _construct_widgets(self, **kwargs):  # pragma: no cover
        pass

    @abstractmethod
    def update(self, val):  # pragma: no cover
        pass

    @abstractmethod
    def recalculate(self, event):  # pragma: no cover
        pass

    def toa(self, event):  # pragma: no cover
        warnings.warn(
            "This function was not implemented for this Phaseogram. "
            "Try the basic one.",
            AstropyUserWarning,
        )

    def reset(self, event):
        for s in self.sliders:
            s.reset()
        self.pcolor.set_array(self.phaseogr.T.ravel())
        self._set_lines(False)
        prof = np.sum(np.nan_to_num(self.unnorm_phaseogr), axis=1)
        ph = np.linspace(0, 2, len(prof) + 1)[:-1]
        self.profile.set_ydata(prof)
        self.proftext.set_text(get_H_label(ph, prof))

    def zoom_in(self, event):
        for s in self.sliders:
            valinit = s.val
            valrange = s.valmax - s.valmin
            valmin = valinit - valrange / 4
            if s.hardvalmin is not None:
                valmin = max(s.hardvalmin, valmin)
            valmax = valinit + valrange / 4
            if s.hardvalmax is not None:
                valmax = min(s.hardvalmax, valmax)
            label = s.label.get_text()
            ax = s.ax
            ax.clear()

            s.__init__(
                ax,
                label,
                valmin=valmin,
                valmax=valmax,
                valinit=valinit,
                hardvalmax=s.hardvalmax,
                hardvalmin=s.hardvalmin,
            )
            ax.text(
                0,
                0,
                str(valmin),
                transform=ax.transAxes,
                horizontalalignment="left",
                color="white",
            )
            ax.text(
                1,
                0,
                str(valmax),
                transform=ax.transAxes,
                horizontalalignment="right",
                color="white",
            )
            s.on_changed(self.update)

    def zoom_out(self, event):
        for s in self.sliders:
            valinit = s.val
            valrange = s.valmax - s.valmin
            valmin = valinit - valrange
            if s.hardvalmin is not None:
                valmin = max(s.hardvalmin, valmin)
            valmax = valinit + valrange
            if s.hardvalmax is not None:
                valmax = min(s.hardvalmax, valmax)
            label = s.label.get_text()
            ax = s.ax
            ax.clear()

            s.__init__(
                ax,
                label,
                valmin=valmin,
                valmax=valmax,
                valinit=valinit,
                hardvalmax=s.hardvalmax,
                hardvalmin=s.hardvalmin,
            )
            ax.text(
                0,
                0,
                str(valmin),
                transform=ax.transAxes,
                horizontalalignment="left",
                color="white",
            )
            ax.text(
                1,
                0,
                str(valmax),
                transform=ax.transAxes,
                horizontalalignment="right",
                color="white",
            )
            s.on_changed(self.update)

    @abstractmethod
    def quit(self, event):  # pragma: no cover
        pass

    @abstractmethod
    def get_values(self):  # pragma: no cover
        pass

    @abstractmethod
    def _line_delay_fun(self, times):  # pragma: no cover
        pass

    @abstractmethod
    def _delay_fun(self, times):  # pragma: no cover
        """This is the delay function _without_ frequency derivatives."""
        pass

    @abstractmethod
    def _read_sliders(self):  # pragma: no cover
        pass

    def _set_lines(self, apply_delay=True):
        if apply_delay:
            func = self._line_delay_fun
        else:

            def func(x):
                return 0

        for i, ph0 in enumerate(self.line_phases):
            linephase = ph0 + func(self.times) - func(self.times[0])
            self.lines[i].set_xdata(linephase)

    def get_timing_model_string(self):

        tm_string = ""

        if self.mjdref is not None:
            tm_string += "PEPOCH         {}\n".format(
                self.pepoch / 86400 + self.mjdref
            )
        tm_string += "PSRJ           {}\n".format(self.object)
        if self.position is not None:
            tm_string += "RAJ            {}\n".format(
                self.position.ra.to_string("hour", sep=":")
            )
            tm_string += "DECJ           {}\n".format(
                self.position.dec.to_string(sep=":")
            )

        tm_string += "F0             {}\n".format(self.freq)
        tm_string += "F1             {}\n".format(self.fdot)
        tm_string += "F2             {}\n".format(self.fddot)

        if hasattr(self, "orbital_period") and self.orbital_period is not None:
            tm_string += "BINARY BT\n"
            tm_string += "PB             {}\n".format(
                self.orbital_period / 86400
            )
            tm_string += "A1             {}\n".format(self.asini)
            if self.mjdref is not None:
                tm_string += "T0             {}\n".format(
                    self.t0 / 86400 + self.mjdref
                )
            tm_string += "T0(MET)        {}\n".format(self.t0)
            tm_string += "PB(s)          {}\n".format(self.orbital_period)

        tm_string += "# PEPOCH(MET)  {}\n".format(self.pepoch)
        start, stop = self.gti.min(), self.gti.max()
        tm_string += "START         {}\n".format(start / 86400 + self.mjdref)
        tm_string += "FINISH        {}\n".format(stop / 86400 + self.mjdref)
        return tm_string


class InteractivePhaseogram(BasePhaseogram):
    def _construct_widgets(self):
        self.df = 0
        self.dfdot = 0
        self.dfddot = 0

        tseg = np.median(np.diff(self.times))
        tobs = tseg * self.nt
        delta_df_start = 4 / tobs
        self.df_order_of_mag = int(np.log10(delta_df_start))
        delta_df = delta_df_start / 10 ** self.df_order_of_mag

        delta_dfdot_start = 8 / tobs ** 2
        self.dfdot_order_of_mag = int(np.log10(delta_dfdot_start))
        delta_dfdot = delta_dfdot_start / 10 ** self.dfdot_order_of_mag

        delta_dfddot_start = 16 / tobs ** 3
        self.dfddot_order_of_mag = int(np.log10(delta_dfddot_start))
        delta_dfddot = delta_dfddot_start / 10 ** self.dfddot_order_of_mag

        freq_str = r"$\Delta$ F0" "x$10^{" + f"{self.df_order_of_mag}" + r"}$"
        fdot_str = (
            r"$\Delta$ F1" r"x$10^{" + f"{self.dfdot_order_of_mag}" + r"}$"
        )
        fddot_str = (
            r"$\Delta$ F2" r"x$10^{" + f"{self.dfddot_order_of_mag}" + r"}$"
        )

        self.sfreq = SliderOnSteroids(
            self.slider_axes[0],
            freq_str,
            -delta_df,
            delta_df,
            valinit=self.df,
            hardvalmin=0,
        )
        self.sfdot = SliderOnSteroids(
            self.slider_axes[1],
            fdot_str,
            -delta_dfdot,
            delta_dfdot,
            valinit=self.dfdot,
        )

        self.sfddot = SliderOnSteroids(
            self.slider_axes[2],
            fddot_str,
            -delta_dfddot,
            delta_dfddot,
            valinit=self.dfddot,
        )

        self.sfreq.on_changed(self.update)
        self.sfdot.on_changed(self.update)
        self.sfddot.on_changed(self.update)
        self.sliders = [self.sfreq, self.sfdot, self.sfddot]

    def update(self, val):
        self._set_lines()
        self.fig.canvas.draw_idle()

    def _read_sliders(self):
        fddot = self.sfddot.val * 10 ** self.dfddot_order_of_mag
        fdot = self.sfdot.val * 10 ** self.dfdot_order_of_mag
        freq = self.sfreq.val * 10 ** self.df_order_of_mag
        return freq, fdot, fddot

    def _line_delay_fun(self, times):
        freq, fdot, fddot = self._read_sliders()
        return (
            (times - self.pepoch).astype(np.float64) * freq
            + 0.5 * (times - self.pepoch) ** 2 * fdot
            + 1 / 6 * (times - self.pepoch) ** 3 * fddot
        )

    def _delay_fun(self, times):
        """This is the delay function _without_ frequency derivatives."""
        return 0

    def recalculate(self, event):
        dfreq, dfdot, dfddot = self._read_sliders()
        pepoch = self.pepoch

        self.fddot = self.fddot - dfddot
        self.fdot = self.fdot - dfdot
        self.freq = self.freq - dfreq

        self.unnorm_phaseogr, _, _, _ = normalized_phaseogram(
            None,
            self.ev_times,
            self.freq,
            fdot=self.fdot,
            plot=False,
            nph=self.nph,
            nt=self.nt,
            pepoch=pepoch,
            fddot=self.fddot,
        )
        self.phaseogr, _, _, _ = normalized_phaseogram(
            self.norm,
            self.ev_times,
            self.freq,
            fdot=self.fdot,
            plot=False,
            nph=self.nph,
            nt=self.nt,
            pepoch=pepoch,
            fddot=self.fddot,
        )

        self.reset(1)

        self.fig.canvas.draw()
        self.timing_model_string = self.get_timing_model_string()
        print("------------------------")
        print(self.timing_model_string)
        print("------------------------")

    def toa(self, event):
        self.timing_model_string = self.get_timing_model_string()

        dfreq, dfdot, dfddot = self._read_sliders()
        freqs = [self.freq - dfreq, self.fdot - dfdot, self.fddot - dfddot]
        folding_length = np.median(np.diff(self.times))
        nbin = self.nph
        template_raw = np.sum(np.nan_to_num(self.unnorm_phaseogr), axis=1)
        template = gaussian_filter1d(template_raw, sigma=1)

        if self.nph < 64:
            warnings.warn(
                "TOA calculation is not robust if the "
                "number of bins is < 64. Oversampling."
            )
            nbin = 64
            phases = np.arange(self.nph * 2 + 1) / (self.nph * 2)
            fun = interp1d(
                phases, np.concatenate((template, [template[0]])), kind="cubic"
            )
            new_phases = np.arange(0, nbin * 2) / (nbin * 2)
            template = fun(new_phases)

        template = template[nbin // 2 : nbin // 2 + template.size // 2]

        template = np.roll(template, -np.argmax(template)) / self.nt

        toa, toaerr = get_TOAs_from_events(
            self.ev_times,
            folding_length,
            *freqs,
            gti=self.gti,
            template=template,
            mjdref=self.mjdref,
            nbin=self.nph,
            pepoch=self.pepoch,
            timfile=self.label + ".tim",
            label=self.label[:10],
            quick=self.test,
            position=None,
        )
        toa_corr = toa + self.time_corr_mjd_fun(toa)
        corr_string = ""

        if np.any(toa_corr != toa):
            corr_string = "_corr"

        with open(self.label + corr_string + ".tim", "w") as fobj:
            print("FORMAT 1", file=fobj)
            for t, te in zip(toa_corr, toaerr):
                print("HEN", 0, t, te, "@", file=fobj)

        with open(self.label + ".par", "w") as fobj:
            print(self.timing_model_string, file=fobj)

    def quit(self, event):
        plt.close(self.fig)

    def get_values(self):
        return self.freq, self.fdot, self.fddot


class BinaryPhaseogram(BasePhaseogram):
    def __init__(self, *args, **kwargs):
        """Init BinaryPhaseogram class.

        Parameters
        ----------
        ev_times : array-like
            Event times
        freq : float
            Frequency of pulsation

        Other parameters
        ----------------
        orbital_period : float
            orbital period in seconds
        asini : float
            projected semi-major axis in light-sec
        T0 : float
            passage through the ascending node, in seconds.
        **kwargs : keyword args
            additional arguments to pass to `BasePhaseogram`
        """
        self.orbital_period = None
        self.asini = 0
        self.t0 = None

        try:
            self.orbital_period = kwargs.pop("orbital_period")
            self.asini = kwargs.pop("asini")
            self.t0 = kwargs.pop("t0")
        except KeyError:
            raise RuntimeError("Please specify all binary parameters")

        BasePhaseogram.__init__(self, *args, **kwargs)

    def _construct_widgets(self):
        self.dperiod = 0
        self.dasini = 0
        self.dt0 = 0

        tseg = np.median(np.diff(self.times))
        tobs = tseg * self.nt

        delta_period = tobs * 5
        delta_asini = 5 / self.freq
        delta_t0 = delta_period

        self.speriod = SliderOnSteroids(
            self.slider_axes[0],
            "PB (s)",
            np.max([0, self.orbital_period - delta_period]),
            self.orbital_period + delta_period,
            valinit=self.orbital_period,
            hardvalmin=0,
        )

        self.sasini = SliderOnSteroids(
            self.slider_axes[1],
            "A1 (l-s)",
            0,
            self.asini + delta_asini,
            valinit=self.asini,
            hardvalmin=0,
        )

        self.st0 = SliderOnSteroids(
            self.slider_axes[2],
            "T0 (MET)",
            self.t0 - delta_t0,
            self.t0 + delta_t0,
            valinit=self.t0,
            hardvalmin=0,
        )

        self.speriod.on_changed(self.update)
        self.sasini.on_changed(self.update)
        self.st0.on_changed(self.update)
        self.sliders = [self.speriod, self.sasini, self.st0]

    def update(self, val):
        self._set_lines()
        self.fig.canvas.draw_idle()

    def _read_sliders(self):
        return self.speriod.val, self.sasini.val, self.st0.val

    def _line_delay_fun(self, times):
        orbital_period, asini, t0 = self._read_sliders()

        new_values = asini * np.sin(2 * np.pi * (times - t0) / orbital_period)
        old_values = self.asini * np.sin(
            2 * np.pi * (times - self.t0) / (self.orbital_period)
        )
        return (new_values - old_values) * self.freq

    def _delay_fun(self, times):
        if self.t0 is None:
            self.t0 = self.pepoch
        if self.orbital_period is None:
            self.orbital_period = self.ev_times[-1] - self.ev_times[0]

        return self.asini * np.sin(
            2 * np.pi * (times - self.t0) / self.orbital_period
        )

    def recalculate(self, event):
        self.orbital_period, self.asini, self.t0 = self._read_sliders()

        corrected_times = self.ev_times - self._delay_fun(self.ev_times)

        self.unnorm_phaseogr, _, _, _ = normalized_phaseogram(
            None,
            corrected_times,
            self.freq,
            fdot=self.fdot,
            plot=False,
            nph=self.nph,
            nt=self.nt,
            pepoch=self.pepoch,
            fddot=self.fddot,
        )

        self.phaseogr, _, _, _ = normalized_phaseogram(
            self.norm,
            corrected_times,
            self.freq,
            fdot=self.fdot,
            plot=False,
            nph=self.nph,
            nt=self.nt,
            pepoch=self.pepoch,
            fddot=self.fddot,
        )

        self._set_lines(False)
        self.pcolor.set_array(self.phaseogr.T.ravel())
        self.sasini.valinit = self.asini
        self.speriod.valinit = self.orbital_period
        self.st0.valinit = self.t0
        self.st0.valmin = self.t0 - self.orbital_period
        self.st0.valmax = self.t0 + self.orbital_period
        self.fig.canvas.draw()

        self.timing_model_string = self.get_timing_model_string()

        print("------------------------")
        print(self.timing_model_string)
        print("------------------------")

    def quit(self, event):
        plt.close(self.fig)

    def get_values(self):
        return self.orbital_period, self.asini, self.t0


def run_interactive_phaseogram(
    event_file,
    freq,
    fdot=0,
    fddot=0,
    nbin=64,
    nt=32,
    binary=False,
    test=False,
    binary_parameters=[None, 0, None],
    pepoch=None,
    norm=None,
    plot_only=False,
    deorbit_par=None,
    emin=None,
    emax=None,
):
    from astropy.io.fits import Header
    from astropy.coordinates import SkyCoord

    events = load_events(event_file)
    if emin is not None or emax is not None:
        events, elabel = filter_energy(events, emin, emax)

    position = name = None

    if hasattr(events, "header") and events.header is not None:
        header = Header.fromstring(events.header)

        try:
            position = SkyCoord(
                header["RA_OBJ"],
                header["DEC_OBJ"],
                unit="deg",
                frame=header["RADECSYS"].lower(),
            )
        except (KeyError, AttributeError):
            position = None

        if "OBJECT" in header:
            name = header["OBJECT"]

    pepoch_mjd = pepoch
    if pepoch is None:
        pepoch = events.gti[0, 0]
        # pepoch_mjd = pepoch / 86400 + events.mjdref
    else:
        pepoch = (pepoch_mjd - events.mjdref) * 86400

    if binary:
        ip = BinaryPhaseogram(
            events.time,
            freq,
            nph=nbin,
            nt=nt,
            fdot=fdot,
            test=test,
            fddot=fddot,
            pepoch=pepoch,
            orbital_period=binary_parameters[0],
            asini=binary_parameters[1],
            t0=binary_parameters[2],
            mjdref=events.mjdref,
            gti=events.gti,
            label=hen_root(event_file),
            norm=norm,
            object=name,
            position=position,
            plot_only=plot_only,
        )
    else:
        events_save = copy.deepcopy(events)
        if deorbit_par is not None:
            events = deorbit_events(events, deorbit_par)

        ip = InteractivePhaseogram(
            events.time,
            freq,
            nph=nbin,
            nt=nt,
            fdot=fdot,
            test=test,
            fddot=fddot,
            pepoch=pepoch,
            mjdref=events.mjdref,
            gti=events.gti,
            label=hen_root(event_file),
            norm=norm,
            object=name,
            position=position,
            plot_only=plot_only,
            time_corr=events_save.time - events.time,
        )

    return ip


def main_phaseogram(args=None):
    description = "Plot an interactive phaseogram"
    from .base import _add_default_args, check_negative_numbers_in_args

    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("file", help="Input event file", type=str)
    parser.add_argument(
        "-f",
        "--freq",
        type=float,
        required=False,
        help="Initial frequency to fold",
        default=None,
    )
    parser.add_argument(
        "--fdot", type=float, required=False, help="Initial fdot", default=0
    )
    parser.add_argument(
        "--fddot", type=float, required=False, help="Initial fddot", default=0
    )
    parser.add_argument(
        "--periodogram",
        type=str,
        required=False,
        help="Periodogram file",
        default=None,
    )
    parser.add_argument(
        "-n",
        "--nbin",
        default=128,
        type=int,
        help="Number of phase bins (X axis) of the profile",
    )
    parser.add_argument(
        "--ntimes",
        default=64,
        type=int,
        help="Number of time bins (Y axis) of the phaseogram",
    )
    parser.add_argument(
        "--binary",
        help="Interact on binary parameters "
        "instead of frequency derivatives",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--binary-parameters",
        help="Initial values for binary parameters",
        default=[None, 0, None],
        nargs=3,
        type=float,
    )
    parser.add_argument(
        "--emin",
        default=None,
        type=float,
        help="Minimum energy (or PI if uncalibrated) to plot",
    )
    parser.add_argument(
        "--emax",
        default=None,
        type=float,
        help="Maximum energy (or PI if uncalibrated) to plot",
    )
    parser.add_argument(
        "--norm",
        help=(
            "Normalization for the phaseogram. Can be 'to1' "
            "(each profile normalized from 0 to 1); "
            "'mediansub' (just subtract the median from each "
            "profile); default None"
        ),
        default=None,
        type=str,
    )
    parser.add_argument(
        "--plot-only",
        help="Only plot the phaseogram",
        default=False,
        action="store_true",
    )

    _add_default_args(
        parser, ["pepoch", "deorbit", "test", "loglevel", "debug"]
    )

    args = check_negative_numbers_in_args(args)
    args = parser.parse_args(args)

    if args.debug:
        args.loglevel = "DEBUG"

    log.setLevel(args.loglevel)

    with log.log_to_file("HENphaseogram.log"):
        if args.periodogram is None and args.freq is None:
            raise ValueError(
                "One of -f or --periodogram arguments MUST be " "specified"
            )
        elif args.periodogram is not None:
            periodogram = load_folding(args.periodogram)
            frequency = float(periodogram.peaks[0])
            fdot = 0
            fddot = 0
        else:
            frequency = args.freq
            fdot = args.fdot
            fddot = args.fddot

        ip = run_interactive_phaseogram(
            args.file,
            freq=frequency,
            fdot=fdot,
            fddot=fddot,
            nbin=args.nbin,
            nt=args.ntimes,
            test=args.test,
            binary=args.binary,
            binary_parameters=args.binary_parameters,
            pepoch=args.pepoch,
            norm=args.norm,
            plot_only=args.plot_only,
            deorbit_par=args.deorbit_par,
            emin=args.emin,
            emax=args.emax,
        )
    plt.close(ip.fig)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Functions to calculate frequency spectra."""

import copy
import warnings
import contextlib
import os
from stingray.gti import cross_gtis
from stingray.crossspectrum import AveragedCrossspectrum
from stingray.powerspectrum import AveragedPowerspectrum
from stingray.utils import show_progress
from stingray.gti import time_intervals_from_gtis
from stingray.events import EventList
import numpy as np
from astropy import log
from astropy.logger import AstropyUserWarning
from .base import (
    hen_root,
    common_name,
    _assign_value_if_none,
    interpret_bintime,
)
from .io import sort_files, save_pds, load_data
from .io import HEN_FILE_EXTENSION, get_file_type


def average_periodograms(fspec_iterable, total=None):
    """Sum a list (or iterable) of power density spectra.

    Examples
    --------
    >>> pds = AveragedPowerspectrum()
    >>> pds.freq = np.asarray([1, 2, 3])
    >>> pds.power = np.asarray([3, 3, 3])
    >>> pds.power_err = np.asarray([0.1, 0.1, 0.1])
    >>> pds.m = 1
    >>> pds.fftlen = 128
    >>> pds1 = copy.deepcopy(pds)
    >>> pds1.m = 2
    >>> tot_pds = average_periodograms([pds, pds1])
    >>> np.allclose(tot_pds.power, pds.power)
    True
    >>> np.allclose(tot_pds.power_err, pds.power_err / np.sqrt(3))
    True
    >>> tot_pds.m
    3
    """

    for i, contents in enumerate(show_progress(fspec_iterable, total=total)):
        freq = contents.freq
        pds = contents.power
        epds = contents.power_err
        nchunks = contents.m
        rebin = 1
        norm = contents.norm
        fftlen = contents.fftlen
        if i == 0:
            rebin0, norm0, freq0 = rebin, norm, freq
            tot_pds = pds * nchunks
            tot_epds = epds ** 2 * nchunks
            tot_npds = nchunks
            tot_contents = copy.copy(contents)
        else:
            assert np.all(
                rebin == rebin0
            ), "Files must be rebinned in the same way"
            np.testing.assert_array_almost_equal(
                freq,
                freq0,
                decimal=int(-np.log10(1 / fftlen) + 2),
                err_msg="Frequencies must coincide",
            )
            assert norm == norm0, "Files must have the same normalization"

            tot_pds += pds * nchunks
            tot_epds += epds ** 2 * nchunks
            tot_npds += nchunks

    tot_contents.power = tot_pds / tot_npds
    tot_contents.power_err = np.sqrt(tot_epds) / tot_npds
    tot_contents.m = tot_npds

    return tot_contents


def _wrap_fun_cpds(arglist):
    f1, f2, outname, kwargs = arglist
    return calc_cpds(f1, f2, outname=outname, **kwargs)


def _wrap_fun_pds(argdict):
    fname = argdict["fname"]
    argdict.pop("fname")
    return calc_pds(fname, **argdict)


def sync_gtis(lc1, lc2):
    """Sync gtis between light curves or event lists.

    Has to work with new and old versions of stingray.

    Examples
    --------
    >>> from stingray.events import EventList
    >>> from stingray.lightcurve import Lightcurve
    >>> ev1 = EventList(
    ...     time=np.sort(np.random.uniform(1, 10, 3)), gti=[[1, 10]])
    >>> ev2 = EventList(time=np.sort(np.random.uniform(0, 9, 4)), gti=[[0, 9]])
    >>> e1, e2 = sync_gtis(ev1, ev2)
    >>> np.allclose(e1.gti, [[1, 9]])
    True
    >>> np.allclose(e2.gti, [[1, 9]])
    True
    >>> lc1 = Lightcurve(
    ...     time=[0.5, 1.5, 2.5], counts=[2, 2, 3], dt=1, gti=[[0, 3]])
    >>> lc2 = Lightcurve(
    ...     time=[1.5, 2.5, 3.5, 4.5], counts=[2, 2, 3, 3], dt=1, gti=[[1, 5]])
    >>> lc1._apply_gtis = lc1.apply_gtis
    >>> lc2._apply_gtis = lc2.apply_gtis
    >>> l1, l2 = sync_gtis(lc1, lc2)
    >>> np.allclose(l1.gti, [[1, 3]])
    True
    >>> np.allclose(l2.gti, [[1, 3]])
    True
    """
    gti = cross_gtis([lc1.gti, lc2.gti])
    lc1.gti = gti
    lc2.gti = gti
    if hasattr(lc1, "_apply_gtis"):
        # Compatibility with old versions of stingray
        lc1.apply_gtis = lc1._apply_gtis
        lc2.apply_gtis = lc2._apply_gtis

    if hasattr(lc1, "apply_gtis"):
        lc1.apply_gtis()
        lc2.apply_gtis()

    # compatibility with old versions of stingray
    if hasattr(lc1, "tseg") and lc1.tseg != lc2.tseg:
        lc1.tseg = np.max(lc1.gti) - np.min(lc1.gti)
        lc2.tseg = np.max(lc1.gti) - np.min(lc1.gti)
    return lc1, lc2


def _format_lc_data(data, type, fftlen=512.0, bintime=1.0):
    if type == "events":
        events = data
        gtilength = events.gti[:, 1] - events.gti[:, 0]
        events.gti = events.gti[gtilength >= fftlen]
        lc_data = list(events.to_lc_list(dt=bintime))
    else:
        lc = data
        if bintime > lc.dt:
            lcrebin = np.rint(bintime / lc.dt)
            log.info("Rebinning lcs by a factor %d" % lcrebin)
            lc = lc.rebin(bintime)
            # To fix problem with float128
            lc.counts = lc.counts.astype(float)
        lc_data = lc
    return lc_data


def _distribute_events(events, chunk_length):
    """Split event list in chunks.

    Examples
    --------
    >>> ev = EventList([1, 2, 3, 4, 5, 6], gti=[[0.5, 6.5]])
    >>> ev.pi = np.ones_like(ev.time)
    >>> ev.mjdref = 56780.
    >>> ev_lists = list(_distribute_events(ev, 2))
    >>> np.allclose(ev_lists[0].time, [1, 2])
    True
    >>> np.allclose(ev_lists[1].time, [3, 4])
    True
    >>> np.allclose(ev_lists[2].time, [5, 6])
    True
    >>> np.allclose(ev_lists[0].gti, [[0.5, 2.5]])
    True
    >>> ev_lists[0].mjdref == ev.mjdref
    True
    >>> ev_lists[2].mjdref == ev.mjdref
    True
    >>> np.allclose(ev_lists[1].pi, [1, 1])
    True
    """
    gti = events.gti
    start_times, stop_times = time_intervals_from_gtis(gti, chunk_length)
    for start, end in zip(start_times, stop_times):
        first, last = np.searchsorted(events.time, [start, end])
        new_ev = EventList(
            events.time[first:last], gti=np.asarray([[start, end]])
        )
        for attr in events.__dict__.keys():
            if attr == "gti":
                continue
            val = getattr(events, attr)
            if np.size(val) == np.size(events.time):
                val = val[first:last]
            setattr(new_ev, attr, val)
        yield new_ev


def _provide_periodograms(events, fftlen, dt, norm):
    for new_ev in _distribute_events(events, fftlen):
        # Hack: epsilon slightly below zero, to allow for a GTI to be recognized as such
        new_ev.gti[:, 1] += dt / 10
        pds = AveragedPowerspectrum(
            new_ev, dt=dt, segment_size=fftlen, norm=norm, silent=True
        )
        pds.fftlen = fftlen
        yield pds


def _provide_cross_periodograms(events1, events2, fftlen, dt, norm):
    length = events1.gti[-1, 1] - events1.gti[0, 0]
    total = int(length / fftlen)
    ev1_iter = _distribute_events(events1, fftlen)
    ev2_iter = _distribute_events(events2, fftlen)
    for new_ev in zip(ev1_iter, ev2_iter):
        new_ev1, new_ev2 = new_ev
        new_ev1.gti[:, 1] += dt / 10
        new_ev2.gti[:, 1] += dt / 10
        with contextlib.redirect_stdout(os.devnull):
            pds = AveragedCrossspectrum(
                new_ev1,
                new_ev2,
                dt=dt,
                segment_size=fftlen,
                norm=norm,
                silent=True,
            )
        pds.fftlen = fftlen
        yield pds


def calc_pds(
    lcfile,
    fftlen,
    save_dyn=False,
    bintime=1,
    pdsrebin=1,
    normalization="leahy",
    back_ctrate=0.0,
    noclobber=False,
    outname=None,
    save_all=False,
    test=False,
):
    """Calculate the PDS from an input light curve file.

    Parameters
    ----------
    lcfile : str
        The light curve file
    fftlen : float
        The length of the chunks over which FFTs will be calculated, in seconds

    Other Parameters
    ----------------
    save_dyn : bool
        If True, save the dynamical power spectrum
    bintime : float
        The bin time. If different from that of the light curve, a rebinning is
        performed
    pdsrebin : int
        Rebin the PDS of this factor.
    normalization: str
        'Leahy', 'frac', 'rms', or any normalization accepted by ``stingray``.
        Default 'Leahy'
    back_ctrate : float
        The non-source count rate
    noclobber : bool
        If True, do not overwrite existing files
    outname : str
        If speficied, output file name. If not specified or None, the new file
        will have the same root as the input light curve and the '_pds' suffix
    """
    root = hen_root(lcfile)
    if outname is None:
        outname = root + "_pds" + HEN_FILE_EXTENSION
    if noclobber and os.path.exists(outname):
        warnings.warn("File exists, and noclobber option used. Skipping")
        return

    ftype, data = get_file_type(lcfile)
    mjdref = data.mjdref
    instr = data.instr

    length = data.gti[-1, 1] - data.gti[0, 0]
    if hasattr(data, "dt"):
        bintime = max(data.dt, bintime)

    nbins = int(length / bintime)

    if ftype == "events" and (test or nbins > 10 ** 7):
        print("Long observation. Using split analysis")
        length = data.gti[-1, 1] - data.gti[0, 0]
        total = int(length / fftlen)
        pds = average_periodograms(
            _provide_periodograms(
                data, fftlen, bintime, norm=normalization.lower()
            ),
            total=total,
        )
    else:
        lc_data = _format_lc_data(data, ftype, bintime=bintime, fftlen=fftlen)

        pds = AveragedPowerspectrum(
            lc_data, segment_size=fftlen, norm=normalization.lower()
        )

    if pdsrebin is not None and pdsrebin != 1:
        pds = pds.rebin(pdsrebin)

    pds.instr = instr
    pds.fftlen = fftlen
    pds.back_phots = back_ctrate * fftlen
    pds.mjdref = mjdref

    log.info("Saving PDS to %s" % outname)
    save_pds(pds, outname, save_all=save_all)
    return outname


def calc_cpds(
    lcfile1,
    lcfile2,
    fftlen,
    save_dyn=False,
    bintime=1,
    pdsrebin=1,
    outname="cpds" + HEN_FILE_EXTENSION,
    normalization="leahy",
    back_ctrate=0.0,
    noclobber=False,
    save_all=False,
    test=False,
):
    """Calculate the CPDS from a pair of input light curve files.

    Parameters
    ----------
    lcfile1 : str
        The first light curve file
    lcfile2 : str
        The second light curve file
    fftlen : float
        The length of the chunks over which FFTs will be calculated, in seconds

    Other Parameters
    ----------------
    save_dyn : bool
        If True, save the dynamical power spectrum
    bintime : float
        The bin time. If different from that of the light curve, a rebinning is
        performed
    pdsrebin : int
        Rebin the PDS of this factor.
    normalization : str
        'Leahy', 'frac', 'rms', or any normalization accepted by ``stingray``.
        Default 'Leahy'
    back_ctrate : float
        The non-source count rate
    noclobber : bool
        If True, do not overwrite existing files
    outname : str
        Output file name for the cpds. Default: cpds.[nc|p]
    """
    if noclobber and os.path.exists(outname):
        warnings.warn("File exists, and noclobber option used. Skipping")
        return

    log.info("Loading file %s..." % lcfile1)
    ftype1, lc1 = get_file_type(lcfile1)
    log.info("Loading file %s..." % lcfile2)
    ftype2, lc2 = get_file_type(lcfile2)
    instr1 = lc1.instr
    instr2 = lc2.instr

    if ftype1 != ftype2:
        raise ValueError(
            "Please use similar data files for the two time "
            "series (e.g. both events or both light curves)"
        )
    if hasattr(lc1, "dt"):
        assert lc1.dt == lc2.dt, "Light curves are sampled differently"

    lc1, lc2 = sync_gtis(lc1, lc2)
    if lc1.mjdref != lc2.mjdref:
        lc2 = lc2.change_mjdref(lc1.mjdref)
    mjdref = lc1.mjdref

    length = lc1.gti[-1, 1] - lc1.gti[0, 0]
    if hasattr(lc1, "dt"):
        bintime = max(lc1.dt, bintime)

    nbins = int(length / bintime)

    if ftype1 == "events" and (test or nbins > 10 ** 7):
        print("Long observation. Using split analysis")
        length = lc1.gti[-1, 1] - lc1.gti[0, 0]
        total = int(length / fftlen)
        cpds = average_periodograms(
            _provide_cross_periodograms(
                lc1, lc2, fftlen, bintime, norm=normalization.lower()
            ),
            total=total,
        )
    else:
        lc1 = _format_lc_data(lc1, ftype1, fftlen=fftlen, bintime=bintime)
        lc2 = _format_lc_data(lc2, ftype2, fftlen=fftlen, bintime=bintime)

        cpds = AveragedCrossspectrum(
            lc1, lc2, segment_size=fftlen, norm=normalization.lower()
        )

    if pdsrebin is not None and pdsrebin != 1:
        cpds = cpds.rebin(pdsrebin)

    cpds.instrs = instr1 + "," + instr2
    cpds.fftlen = fftlen
    cpds.back_phots = back_ctrate * fftlen
    cpds.mjdref = mjdref
    lags, lags_err = cpds.time_lag()
    cpds.lag = lags
    cpds.lag_err = lags

    log.info("Saving CPDS to %s" % outname)
    save_pds(cpds, outname, save_all=save_all)
    return outname


def calc_fspec(
    files,
    fftlen,
    do_calc_pds=True,
    do_calc_cpds=True,
    do_calc_cospectrum=True,
    do_calc_lags=True,
    save_dyn=False,
    bintime=1,
    pdsrebin=1,
    outroot=None,
    normalization="leahy",
    nproc=1,
    back_ctrate=0.0,
    noclobber=False,
    ignore_instr=False,
    save_all=False,
    test=False,
):
    r"""Calculate the frequency spectra: the PDS, the cospectrum, ...

    Parameters
    ----------
    files : list of str
        List of input file names
    fftlen : float
        length of chunks to perform the FFT on.

    Other Parameters
    ----------------
    save_dyn : bool
        If True, save the dynamical power spectrum
    bintime : float
        The bin time. If different from that of the light curve, a rebinning is
        performed
    pdsrebin : int
        Rebin the PDS of this factor.
    normalization : str
        'Leahy' [3] or 'rms' [4] [5]. Default 'Leahy'.
    back_ctrate : float
        The non-source count rate
    noclobber : bool
        If True, do not overwrite existing files
    outroot : str
        Output file name root
    nproc : int
        Number of processors to use to parallelize the processing of multiple
        files
    ignore_instr : bool
        Ignore instruments; files are alternated in the two channels

    References
    ----------
    [3] Leahy et al. 1983, ApJ, 266, 160.

    [4] Belloni & Hasinger 1990, A&A, 230, 103

    [5] Miyamoto et al. 1991, ApJ, 383, 784

    """

    log.info("Using %s normalization" % normalization)
    log.info("Using %s processors" % nproc)

    if do_calc_pds:
        wrapped_file_dicts = []
        for f in files:
            wfd = dict(
                fftlen=fftlen,
                save_dyn=save_dyn,
                bintime=bintime,
                pdsrebin=pdsrebin,
                normalization=normalization.lower(),
                back_ctrate=back_ctrate,
                noclobber=noclobber,
                save_all=save_all,
                test=test,
            )
            wfd["fname"] = f
            wrapped_file_dicts.append(wfd)

        [_wrap_fun_pds(w) for w in wrapped_file_dicts]

    if not do_calc_cpds or len(files) < 2:
        return

    if ignore_instr:
        files1 = files[0::2]
        files2 = files[1::2]
    else:
        log.info("Sorting file list")
        sorted_files = sort_files(files)

        warnings.warn(
            "Beware! For cpds and derivatives, I assume that the "
            "files are from only two instruments and in pairs "
            "(even in random order)"
        )

        instrs = list(sorted_files.keys())

        files1 = sorted_files[instrs[0]]
        files2 = sorted_files[instrs[1]]

    assert len(files1) == len(files2), "An even number of files is needed"

    argdict = dict(
        fftlen=fftlen,
        save_dyn=save_dyn,
        bintime=bintime,
        pdsrebin=pdsrebin,
        normalization=normalization.lower(),
        back_ctrate=back_ctrate,
        noclobber=noclobber,
        save_all=save_all,
        test=test,
    )

    funcargs = []

    for i_f, f in enumerate(files1):
        f1, f2 = f, files2[i_f]

        outdir = os.path.dirname(f1)
        if outdir == "":
            outdir = os.getcwd()

        outr = _assign_value_if_none(
            outroot, common_name(f1, f2, default="%d" % i_f)
        )

        outname = os.path.join(
            outdir,
            outr.replace(HEN_FILE_EXTENSION, "")
            + "_cpds"
            + HEN_FILE_EXTENSION,
        )

        funcargs.append([f1, f2, outname, argdict])

    [_wrap_fun_cpds(fa) for fa in funcargs]


def _normalize(array, ref=0):
    """Normalize array in terms of standard deviation.

    Examples
    --------
    >>> n = 10000
    >>> array1 = np.random.normal(0, 1, n)
    >>> array2 = np.random.normal(0, 1, n)
    >>> array = array1 ** 2 + array2 ** 2
    >>> newarr = _normalize(array)
    >>> np.isclose(np.std(newarr), 1, atol=0.0001)
    True
    """
    m = ref
    std = np.std(array)
    newarr = np.zeros_like(array)
    good = array > m
    newarr[good] = (array[good] - ref) / std
    return newarr


def dumpdyn(fname, plot=False):
    raise NotImplementedError(
        "Dynamical power spectrum is being refactored. "
        "Sorry for the inconvenience. In the meantime, "
        "you can load the data into Stingray using "
        "`cs = hendrics.io.load_pds(fname)` and find "
        "the dynamical PDS/CPDS in cs.cs_all"
    )


def dumpdyn_main(args=None):
    """Main function called by the `HENdumpdyn` command line script."""
    import argparse

    description = (
        "Dump dynamical (cross) power spectra. "
        "This script is being reimplemented. Please be "
        "patient :)"
    )
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument(
        "files",
        help=("List of files in any valid HENDRICS " "format for PDS or CPDS"),
        nargs="+",
    )
    parser.add_argument(
        "--noplot", help="plot results", default=False, action="store_true"
    )

    args = parser.parse_args(args)

    fnames = args.files

    for f in fnames:
        dumpdyn(f, plot=not args.noplot)


def main(args=None):
    """Main function called by the `HENfspec` command line script."""
    import argparse
    from .base import _add_default_args, check_negative_numbers_in_args

    description = (
        "Create frequency spectra (PDS, CPDS, cospectrum) "
        "starting from well-defined input ligthcurves"
    )
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of light curve files", nargs="+")
    parser.add_argument(
        "-b",
        "--bintime",
        type=float,
        default=1 / 4096,
        help="Light curve bin time; if negative, interpreted"
        + " as negative power of 2."
        + " Default: 2^-10, or keep input lc bin time"
        + " (whatever is larger)",
    )
    parser.add_argument(
        "-r",
        "--rebin",
        type=int,
        default=1,
        help="(C)PDS rebinning to apply. Default: none",
    )
    parser.add_argument(
        "-f",
        "--fftlen",
        type=float,
        default=512,
        help="Length of FFTs. Default: 512 s",
    )
    parser.add_argument(
        "-k",
        "--kind",
        type=str,
        default="PDS,CPDS,cos",
        help="Spectra to calculate, as comma-separated list"
        + " (Accepted: PDS and CPDS;"
        + ' Default: "PDS,CPDS")',
    )
    parser.add_argument(
        "--norm",
        type=str,
        default="leahy",
        help="Normalization to use"
        + " (Accepted: leahy and rms;"
        + ' Default: "leahy")',
    )
    parser.add_argument(
        "--noclobber",
        help="Do not overwrite existing files",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "-o",
        "--outroot",
        type=str,
        default=None,
        help="Root of output file names for CPDS only",
    )
    parser.add_argument(
        "--back",
        help=("Estimated background (non-source) count rate"),
        default=0.0,
        type=float,
    )
    parser.add_argument(
        "--save-dyn",
        help="save dynamical power spectrum",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--ignore-instr",
        help="Ignore instrument names in channels",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--save-all",
        help="Save all information contained in spectra,"
        " including single pdss and light curves.",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--test",
        help="Only to be used in testing",
        default=False,
        action="store_true",
    )
    _add_default_args(parser, ["loglevel", "debug"])

    args = check_negative_numbers_in_args(args)
    args = parser.parse_args(args)

    if args.debug:
        args.loglevel = "DEBUG"

    log.setLevel(args.loglevel)

    with log.log_to_file("HENfspec.log"):
        bintime = np.longdouble(interpret_bintime(args.bintime))

        fftlen = args.fftlen
        pdsrebin = args.rebin
        normalization = args.norm
        if normalization.lower() not in [
            "frac",
            "abs",
            "leahy",
            "none",
            "rms",
        ]:
            warnings.warn("Beware! Unknown normalization!", AstropyUserWarning)
            normalization = "leahy"
        if normalization == "rms":
            normalization = "frac"

        do_cpds = do_pds = do_cos = do_lag = False
        kinds = args.kind.split(",")
        for k in kinds:
            if k == "PDS":
                do_pds = True
            elif k == "CPDS":
                do_cpds = True
            elif k == "cos" or k == "cospectrum":
                do_cos = True
                do_cpds = True
            elif k == "lag":
                do_lag = True
                do_cpds = True

        calc_fspec(
            args.files,
            fftlen,
            do_calc_pds=do_pds,
            do_calc_cpds=do_cpds,
            do_calc_cospectrum=do_cos,
            do_calc_lags=do_lag,
            save_dyn=args.save_dyn,
            bintime=bintime,
            pdsrebin=pdsrebin,
            outroot=args.outroot,
            normalization=normalization,
            nproc=1,
            back_ctrate=args.back,
            noclobber=args.noclobber,
            ignore_instr=args.ignore_instr,
            save_all=args.save_all,
            test=args.test,
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""This is proposed as an Astropy affiliated package."""

# Affiliated packages may add whatever they like to this file, but
# should keep this content at the top.
# ----------------------------------------------------------------------------
from ._astropy_init import *

# ----------------------------------------------------------------------------

# For egg_info test builds to pass, put package imports here.
if not _ASTROPY_SETUP_:
    # Workaround: import netCDF4 before everything else. This loads the HDF5
    # library that netCDF4 uses and not something else.

    try:
        import netCDF4 as nc

        HEN_FILE_EXTENSION = ".nc"
        HAS_NETCDF = True
    except ImportError:
        HEN_FILE_EXTENSION = ".p"
        HAS_NETCDF = False
        pass

    import stingray
    import warnings

    warnings.filterwarnings("ignore", message=".*Errorbars on cross.*")

    from .compat import (
        prange,
        array_take,
        HAS_NUMBA,
        njit,
        vectorize,
        float32,
        float64,
        int32,
        int64,
        _MonkeyPatchedEventList,
    )

    stingray.events.EventList = _MonkeyPatchedEventList



#!/usr/bin/env python

import os
import argparse
import warnings
import numpy as np
import matplotlib.pyplot as plt
import astropy.io.fits as pf
from astropy import log
from astropy.logger import AstropyUserWarning

from stingray.io import load_events_and_gtis, ref_mjd
from stingray.pulse.pulsar import pulse_phase, phase_exposure
from .io import is_string, save_as_qdp
from .base import _assign_value_if_none, hen_root
from .fold import fit_profile, std_fold_fit_func


def outfile_name(file):
    """Output file name for phasetag.

    Examples
    --------
    >>> outfile_name('file.s.a.fits.Z')
    'file.s.a_phasetag.fits.Z'
    >>> outfile_name('file.s.a.evct.gz')
    'file.s.a_phasetag.evct.gz'
    >>> outfile_name('file.s.a.evct')
    'file.s.a_phasetag.evct'
    """
    root, ext = os.path.splitext(file)
    if ext.lower() in [".gz", ".z"]:
        root, newext = os.path.splitext(root)
        ext = newext + ext

    return root + "_phasetag" + ext


def phase_tag(
    ev_list,
    parameter_info,
    gtis=None,
    mjdref=0,
    nbin=10,
    ref_to_max=False,
    pepoch=None,
    expocorr=True,
    pulse_ref_time=None,
    plot=True,
    test=False,
):
    """Phase-tag events in a FITS file with a given ephemeris.

    Parameters
    ----------
    ev_list : float
        Event times
    parameter_info : str or array of floats
        If a string, this is a pulsar parameter file that PINT is able to
        understand. Otherwise, this is a list of frequency derivatives
        [F0, F1, F2, ...]

    Other parameters
    ----------------
    gtis : [[g0_0, g0_1], [g1_0, g1_1], ...]
        Good time intervals
    nbin : int
        Number of nbin in the pulsed profile
    ref_to_max : bool
        Automatically refer the TOAs to the maximum of the profile
    pepoch : float, default None
        Reference epoch for the timing solution. If None, this is the start
        of the observation.
    pulse_ref_time : float
        Reference time for the pulse. This overrides ref_to_max
    plot : bool
        Plot diagnostics
    expocorr : bool
        Use exposure correction when calculating the profile

    """
    # ---- in MJD ----
    if gtis is None:
        gtis = np.array([[ev_list[0], ev_list[-1]]])

    ev_mjd = ev_list / 86400 + mjdref
    gtis_mjd = gtis / 86400 + mjdref

    pepoch = _assign_value_if_none(pepoch, gtis_mjd[0, 0])

    # ------ Orbital DEMODULATION --------------------
    if is_string(parameter_info):
        raise NotImplementedError(
            "This part is not yet implemented. Please "
            "use single frequencies and pepoch as "
            "documented"
        )

    else:
        frequency_derivatives = parameter_info
        times = (ev_mjd - pepoch) * 86400

    f = frequency_derivatives[0]

    phase = pulse_phase(times, *frequency_derivatives, to_1=False)
    gti_phases = pulse_phase(
        (gtis_mjd - pepoch) * 86400, *frequency_derivatives, to_1=False
    )

    # ------- now apply period derivatives ------

    log.info("Calculating phases...")
    ref_phase = 0
    ref_time = 0

    if pulse_ref_time is not None:
        ref_time = (pulse_ref_time - pepoch) * 86400
        ref_phase = ref_time * f
    elif ref_to_max:
        phase_to1 = phase - np.floor(phase)

        raw_profile, bins = np.histogram(
            phase_to1, bins=np.linspace(0, 1, nbin + 1)
        )
        exposure = phase_exposure(
            gti_phases[0, 0], gti_phases[-1, 1], 1, nbin=nbin, gtis=gti_phases
        )
        profile = raw_profile / exposure
        profile_err = np.sqrt(raw_profile) / exposure

        sinpars, bu, bu = fit_profile(
            profile, profile_err, nperiods=2, baseline=True, debug=test
        )
        fine_phases = np.linspace(0, 2, 1000 * 2)
        fitted_profile = std_fold_fit_func(sinpars, fine_phases)
        maxp = np.argmax(fitted_profile)
        ref_phase = fine_phases[maxp]
        if test:  # pragma: no cover
            # No tests with a pulsed profile yet
            ref_phase = bins[np.argmax(raw_profile)]

        ref_time = ref_phase / f

    phase -= ref_phase
    gti_phases -= ref_phase
    phase_to1 = phase - np.floor(phase)

    raw_profile, bins = np.histogram(
        phase_to1, bins=np.linspace(0, 1, nbin + 1)
    )

    exposure = phase_exposure(
        gti_phases[0, 0], gti_phases[-1, 1], 1, nbin=nbin, gtis=gti_phases
    )
    if np.any(np.logical_or(exposure != exposure, exposure == 0)):
        warnings.warn(
            "Exposure has NaNs or zeros. Profile is not normalized",
            AstropyUserWarning,
        )
        expocorr = False

    if not expocorr:
        exposure = np.ones_like(raw_profile)

    profile = raw_profile / exposure
    profile = np.append(profile, profile)
    exposure = np.append(exposure, exposure)
    profile_err = np.sqrt(profile)
    phs = (bins[1:] + bins[:-1]) / 2
    phs = np.append(phs, phs + 1)

    fig = None
    if plot:
        fig = plt.figure()
        plt.errorbar(
            phs, profile / exposure, yerr=profile_err / exposure, fmt="none"
        )
        plt.plot(phs, profile / exposure, "k-", drawstyle="steps-mid")
        plt.xlabel("Phase")
        plt.ylabel("Counts")
        for i in range(20):
            plt.axvline(i * 0.1, ls="--", color="b")
        if not test:  # pragma: no cover
            plt.show()
        else:
            plt.close(fig)

    # ------ WRITE RESULTS BACK TO FITS --------------
    results = type("results", (object,), {})
    results.ev_list = ev_list
    results.phase = phase
    results.frequency_derivatives = frequency_derivatives
    results.ref_time = ref_time
    results.figure = fig
    results.plot_phase = phs
    results.plot_profile = profile / exposure
    results.plot_profile_err = profile_err / exposure
    return results


def phase_tag_fits(
    filename,
    parameter_info,
    gtistring="GTI,STDGTI",
    gti_file=None,
    hduname="EVENTS",
    column="TIME",
    **kwargs
):
    """Phase-tag events in a FITS file with a given ephemeris.

    Parameters
    ----------
    filename : str
        Events FITS file
    parameter_info : str or array of floats
        If a string, this is a pulsar parameter file that PINT is able to
        understand. Otherwise, this is a list of frequency derivatives
        [F0, F1, F2, ...]

    Other parameters
    ----------------
    nbin : int
        Number of nbin in the pulsed profile
    ref_to_max : bool
        Automatically refer the TOAs to the maximum of the profile
    pepoch : float, default None
        Reference epoch for the timing solution. If None, this is the start
        of the observation.
    pulse_ref_time : float
        Reference time for the pulse. This overrides ref_to_max
    plot : bool
        Plot diagnostics
    expocorr : bool
        Use exposure correction when calculating the profile
    gtistring : str
        Comma-separated list of accepted GTI extensions (default GTI,STDGTI),
        with or without appended integer number denoting the detector
    gti_file : str, default None
        External GTI file
    hduname : str, default 'EVENTS'
        Name of the HDU containing the event list
    """

    outfile = outfile_name(filename)
    evreturns = load_events_and_gtis(
        filename,
        gtistring=gtistring,
        gti_file=gti_file,
        hduname=hduname,
        column=column,
    )
    mjdref = ref_mjd(filename)

    results = phase_tag(
        evreturns.ev_list,
        parameter_info,
        gtis=evreturns.gti_list,
        mjdref=mjdref,
        **kwargs
    )
    if results.figure is not None:
        results.figure.savefig(hen_root(filename) + ".pdf")
    phase = results.phase
    frequency_derivatives = results.frequency_derivatives
    ref_time = results.ref_time

    phase_to1 = phase - np.floor(phase)

    # Save results to fits file
    hdulist = pf.open(filename, checksum=True)
    tbhdu = hdulist[hduname]
    table = tbhdu.data
    order = np.argsort(table[column])
    # load_events_and_gtis sorts the data automatically. This is not the case
    # of the other operations done here. So, let's sort the table first.
    for col in table.names:
        table.field(col)[:] = np.array(table[col])[order]

    # If columns already exist, overwrite them. Else, create them
    create = False
    if "Orbit_bary" in table.names:
        table.field("Orbit_bary")[:] = evreturns.ev_list
        table.field("TotPhase_s")[:] = (
            phase / frequency_derivatives[0] + ref_time
        )
        table.field("Phase")[:] = phase_to1
    else:
        create = True

    # first of all, copy columns
    cols = table.columns

    # create new list of columns, copying all columns from other table
    newlist = []

    for c in cols:
        newlist.append(c)

    if create:
        # then, create new column with orbital demodulation
        newcol = pf.Column(
            name="Orbit_bary", format="1D", unit="s", array=results.ev_list
        )

        # append it to new table
        newlist.append(newcol)

        # Do the same with total phase
        newcol = pf.Column(
            name="TotPhase_s",
            format="1D",
            unit="s",
            array=phase / frequency_derivatives[0] + ref_time,
        )

        newlist.append(newcol)

        # Do the same with fractional phase
        newcol = pf.Column(
            name="Phase", format="1D", unit="phase", array=phase_to1
        )

        newlist.append(newcol)

        # make new column definitions
    coldefs = pf.ColDefs(newlist)

    # create new record
    newrec = pf.FITS_rec.from_columns(coldefs)

    # and new hdu
    newtbhdu = pf.BinTableHDU(
        data=newrec, header=tbhdu.header.copy(), name=hduname, uint=False
    )

    # Copy primary HDU from old file
    prihdu = hdulist[0].copy()

    # define new hdulist
    newhdulist = pf.HDUList([prihdu])

    # Copy remaining HDUs from old file
    for h in hdulist[1:]:
        if h.name == hduname:
            newhdulist.append(newtbhdu)
        else:
            newhdulist.append(h.copy())

    newhdulist.writeto(outfile, overwrite=True, checksum=True)
    hdulist.close()

    save_as_qdp(
        [results.plot_phase, results.plot_profile, results.plot_profile_err],
        filename=outfile.replace(".evt", "") + ".qdp",
    )


def main_phasetag(args=None):
    from .base import check_negative_numbers_in_args

    parser = argparse.ArgumentParser()

    parser.add_argument("file", help="Event file", type=str)
    parser.add_argument(
        "--parfile", help="Parameter file", type=str, default=None
    )
    parser.add_argument(
        "-f",
        "--freqs",
        help="Frequency derivatives",
        type=float,
        default=None,
        nargs="+",
    )
    parser.add_argument("-n", "--nbin", type=int, default=16, help="Nbin")
    parser.add_argument(
        "--plot",
        action="store_true",
        default=False,
        dest="plot",
        help="Plot profile",
    )
    parser.add_argument(
        "--tomax",
        action="store_true",
        default=False,
        dest="tomax",
        help="Refer phase to pulse max",
    )
    parser.add_argument(
        "--test",
        action="store_true",
        default=False,
        dest="test",
        help="Only for unit tests! Do not use",
    )
    parser.add_argument(
        "--refTOA",
        default=None,
        type=np.longdouble,
        help="Reference TOA in MJD (overrides --tomax) for "
        "reference pulse phase",
        dest="pulse_ref_time",
    )
    parser.add_argument(
        "--pepoch",
        default=None,
        type=np.longdouble,
        help="Reference time for timing solution",
        dest="pepoch",
    )

    args = check_negative_numbers_in_args(args)
    args = parser.parse_args(args)

    if args.freqs is None and args.parfile is None:
        raise ValueError("Specify one between --freqs or --parfile")
    elif args.freqs is not None and args.parfile is not None:
        raise ValueError("Specify only one between --freqs or --parfile")
    elif args.freqs is not None:
        parameter_info = args.freqs
    else:
        parameter_info = args.parfile

    plot = args.plot
    expocorr = True

    phase_tag_fits(
        args.file,
        parameter_info,
        plot=plot,
        nbin=args.nbin,
        test=args.test,
        expocorr=expocorr,
        ref_to_max=args.tomax,
        pulse_ref_time=args.pulse_ref_time,
        pepoch=args.pepoch,
    )



def _dummy_par(par, pb=1e20, a1=0.0, f0=1.0):
    with open(par, "w") as fobj:
        print("PSRJ     FAKE_X-1", file=fobj)
        print("RAJ      00:55:01", file=fobj)
        print("DECJ     12:00:40.2", file=fobj)
        print("PEPOCH   560000.0", file=fobj)
        print(f"F0       {f0}", file=fobj)
        print("BINARY   BT", file=fobj)
        print("DM       0", file=fobj)
        print(f"PB       {pb}", file=fobj)
        print(f"A1       {a1}", file=fobj)
        print(f"OM       0.0", file=fobj)
        print("ECC      0.0", file=fobj)
        print("T0       56000", file=fobj)
        print("EPHEM    DE421", file=fobj)
        print("CLK      TT(TAI)", file=fobj)

    return par



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from astropy.io.fits import Header
from stingray.events import EventList
from stingray.lightcurve import Lightcurve
from stingray.powerspectrum import Powerspectrum, AveragedPowerspectrum
from stingray.powerspectrum import Crossspectrum, AveragedCrossspectrum
import numpy as np
import os
from hendrics.io import load_events, save_events, save_lcurve, load_lcurve
from hendrics.io import save_data, load_data, save_pds, load_pds
from hendrics.io import HEN_FILE_EXTENSION, _split_high_precision_number
from hendrics.io import save_model, load_model, HAS_C256, HAS_NETCDF
from hendrics.io import find_file_in_allowed_paths
from hendrics.io import save_as_ascii, save_as_qdp, read_header_key, ref_mjd
from hendrics.io import main

import pytest
import glob
from astropy.modeling import models
from astropy.modeling.core import Model

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


def _dummy_bad(x, z, y=0):
    return


def _dummy(x, y=0):
    return


def test_find_files_in_allowed_paths(capsys):

    with open("bu", "w") as fobj:
        print("blabla", file=fobj)
    fakepath = os.path.join("directory", "bu")
    realpath = os.path.join(".", "bu")
    foundpath = find_file_in_allowed_paths(fakepath, ["."])
    stdout, stderr = capsys.readouterr()
    assert "Parfile found at different path" in stdout
    assert foundpath == realpath
    assert find_file_in_allowed_paths("bu") == "bu"
    assert not find_file_in_allowed_paths(os.path.join("directory", "bu"))
    assert not find_file_in_allowed_paths(None)

    os.unlink("bu")


def test_high_precision_keyword():
    """Test high precision FITS keyword read."""
    from hendrics.io import high_precision_keyword_read

    hdr = {
        "MJDTESTI": 100,
        "MJDTESTF": np.longdouble(0.5),
        "CIAO": np.longdouble(0.0),
    }
    assert high_precision_keyword_read(hdr, "MJDTEST") == np.longdouble(100.5)
    assert high_precision_keyword_read(hdr, "CIAO") == np.longdouble(0.0)


class TestIO:
    """Real unit tests."""

    @classmethod
    def setup_class(cls):
        cls.dum = "bubu" + HEN_FILE_EXTENSION
        curdir = os.path.abspath(os.path.dirname(__file__))
        cls.datadir = os.path.join(curdir, "data")

    def test_readfile_event(self):
        fname = os.path.join(self.datadir, "monol_testA.evt")
        main([fname])

    def test_read_header_key(self):
        fname = os.path.join(self.datadir, "monol_testA.evt")
        val = read_header_key(fname, "TSTART")
        assert val == 80000000.0

    def test_ref_mjd(self):
        fname = os.path.join(self.datadir, "monol_testA.evt")
        val = ref_mjd([fname])
        assert np.isclose(val, 55197.00076601852, atol=0.00000000001)

    def test_save_data(self):
        struct = {
            "a": 0.1,
            "b": np.longdouble("123.4567890123456789"),
            "c": np.longdouble([[-0.5, 3.5]]),
            "d": 1,
        }
        save_data(struct, self.dum)
        struct2 = load_data(self.dum)
        assert np.allclose(struct["a"], struct2["a"])
        assert np.allclose(struct["b"], struct2["b"])
        assert np.allclose(struct["c"], struct2["c"])
        assert np.allclose(struct["d"], struct2["d"])

    def test_load_and_save_events(self):
        events = EventList(
            [0, 2, 3.0],
            pi=[1, 2, 3],
            mjdref=54385.3254923845,
            gti=np.longdouble([[-0.5, 3.5]]),
        )
        events.cal_pi = events.pi.copy()
        events.energy = np.array([3.0, 4.0, 5.0])
        events.mission = "nustar"
        events.header = Header().tostring()
        save_events(events, self.dum)
        events2 = load_events(self.dum)
        assert np.allclose(events.time, events2.time)
        assert np.allclose(events.cal_pi, events2.cal_pi)
        assert np.allclose(events.pi, events2.pi)
        assert np.allclose(events.mjdref, events2.mjdref)
        assert np.allclose(events.gti, events2.gti)
        assert np.allclose(events.energy, events2.energy)
        assert events.header == events2.header
        assert events2.mission == events.mission

    def test_load_and_save_lcurve(self):
        lcurve = Lightcurve(
            np.linspace(0, 10, 15),
            np.random.poisson(30, 15),
            mjdref=54385.3254923845,
            gti=np.longdouble([[-0.5, 3.5]]),
        )
        # Monkeypatch for compatibility with old versions
        lcurve.mission = "bububu"
        lcurve.instr = "bababa"
        save_lcurve(lcurve, self.dum)
        lcurve2 = load_lcurve(self.dum)
        assert np.allclose(lcurve.time, lcurve2.time)
        assert np.allclose(lcurve.counts, lcurve2.counts)
        assert np.allclose(lcurve.mjdref, lcurve2.mjdref)
        assert np.allclose(lcurve.gti, lcurve2.gti)
        assert lcurve.err_dist == lcurve2.err_dist
        assert lcurve.mission == lcurve2.mission
        assert lcurve2.mission == "bububu"
        assert lcurve.instr == lcurve2.instr

    def test_load_and_save_pds(self):
        pds = Powerspectrum()
        pds.freq = np.linspace(0, 10, 15)
        pds.power = np.random.poisson(30, 15)
        pds.mjdref = 54385.3254923845
        pds.gti = np.longdouble([[-0.5, 3.5]])
        pds.show_progress = True
        pds.amplitude = False

        save_pds(pds, self.dum)
        pds2 = load_pds(self.dum)
        for attr in ["gti", "mjdref", "m", "show_progress", "amplitude"]:
            assert np.allclose(getattr(pds, attr), getattr(pds2, attr))

    def test_load_and_save_cpds(self):
        pds = Crossspectrum()
        pds.freq = np.linspace(0, 10, 15)
        pds.power = np.random.poisson(30, 15) + 1.0j
        pds.mjdref = 54385.3254923845
        pds.gti = np.longdouble([[-0.5, 3.5]])
        pds.show_progress = True
        pds.amplitude = False

        save_pds(pds, self.dum)
        pds2 = load_pds(self.dum)
        for attr in [
            "gti",
            "mjdref",
            "m",
            "show_progress",
            "amplitude",
            "power",
        ]:
            assert np.allclose(getattr(pds, attr), getattr(pds2, attr))

    def test_load_pds_fails(self):
        pds = EventList()
        save_events(pds, self.dum)
        with pytest.raises(ValueError) as excinfo:
            load_pds(self.dum)
        assert "Unrecognized data type" in str(excinfo.value)

    def test_load_and_save_xps(self):
        lcurve1 = Lightcurve(
            np.linspace(0, 10, 150),
            np.random.poisson(30, 150),
            mjdref=54385.3254923845,
            gti=np.longdouble([[-0.5, 3.5]]),
        )
        lcurve2 = Lightcurve(
            np.linspace(0, 10, 150),
            np.random.poisson(30, 150),
            mjdref=54385.3254923845,
            gti=np.longdouble([[-0.5, 3.5]]),
        )

        xps = AveragedCrossspectrum(lcurve1, lcurve2, 1, save_all=True)
        save_pds(xps, self.dum, save_all=True)
        xps2 = load_pds(self.dum)
        assert np.allclose(xps.gti, xps2.gti)
        assert hasattr(xps2, "cs_all")
        assert np.allclose(xps.cs_all[0].power, xps2.cs_all[0].power)
        assert xps.m == xps2.m
        lag, lag_err = xps.time_lag()
        lag2, lag2_err = xps2.time_lag()
        assert np.allclose(lag, lag2)
        assert hasattr(xps2, "pds1")

    def test_load_and_save_xps_no_all(self):
        lcurve1 = Lightcurve(
            np.linspace(0, 10, 150),
            np.random.poisson(30, 150),
            mjdref=54385.3254923845,
            gti=np.longdouble([[-0.5, 3.5]]),
        )
        lcurve2 = Lightcurve(
            np.linspace(0, 10, 150),
            np.random.poisson(30, 150),
            mjdref=54385.3254923845,
            gti=np.longdouble([[-0.5, 3.5]]),
        )

        xps = AveragedCrossspectrum(lcurve1, lcurve2, 1)

        outfile = "small_xps" + HEN_FILE_EXTENSION
        save_pds(xps, outfile, save_all=False)
        xps2 = load_pds(outfile)
        assert np.allclose(xps.gti, xps2.gti)
        assert xps.m == xps2.m
        assert not hasattr(xps2, "pds1")

    def test_load_and_save_xps_quick(self):
        lcurve1 = Lightcurve(
            np.linspace(0, 10, 150),
            np.random.poisson(30, 150),
            mjdref=54385.3254923845,
            gti=np.longdouble([[-0.5, 3.5]]),
        )
        lcurve2 = Lightcurve(
            np.linspace(0, 10, 150),
            np.random.poisson(30, 150),
            mjdref=54385.3254923845,
            gti=np.longdouble([[-0.5, 3.5]]),
        )

        xps = AveragedCrossspectrum(lcurve1, lcurve2, 1)

        save_pds(xps, self.dum)
        xps2 = load_pds(self.dum, nosub=True)
        assert np.allclose(xps.gti, xps2.gti)
        assert xps.m == xps2.m

        assert not hasattr(xps2, "pds1")

    def test_high_precision_split1(self):
        C_I, C_F, C_l, k = _split_high_precision_number(
            "C", np.double(0.01), 8
        )
        assert C_I == 1
        np.testing.assert_almost_equal(C_F, 0, 6)
        assert C_l == -2
        assert k == "double"

    def test_high_precision_split2(self):
        C_I, C_F, C_l, k = _split_high_precision_number(
            "C", np.double(1.01), 8
        )
        assert C_I == 1
        np.testing.assert_almost_equal(C_F, np.double(0.01), 6)
        assert C_l == 0
        assert k == "double"

    def test_save_longcomplex(self):
        val = np.longcomplex(1.01 + 2.3j)
        data = {"val": val}
        save_data(data, "bubu" + HEN_FILE_EXTENSION)
        data_out = load_data("bubu" + HEN_FILE_EXTENSION)

        assert np.allclose(data["val"], data_out["val"])

    @pytest.mark.skipif("not HAS_C256 or not HAS_NETCDF")
    def test_save_longcomplex(self):
        val = np.complex256(1.01 + 2.3j)
        data = {"val": val}
        with pytest.warns(UserWarning) as record:
            save_data(data, "bubu" + HEN_FILE_EXTENSION)
        assert "complex256 yet unsupported" in record[0].message.args[0]
        data_out = load_data("bubu" + HEN_FILE_EXTENSION)

        assert np.allclose(data["val"], data_out["val"])

    def test_save_as_qdp(self):
        """Test saving arrays in a qdp file."""
        arrays = [np.array([0, 1, 3]), np.array([1, 4, 5])]
        errors = [np.array([1, 1, 1]), np.array([[1, 0.5], [1, 0.5], [1, 1]])]
        save_as_qdp(
            arrays, errors, filename=os.path.join(self.datadir, "bububu.txt")
        )
        save_as_qdp(
            arrays,
            errors,
            filename=os.path.join(self.datadir, "bububu.txt"),
            mode="a",
        )

    def test_save_as_ascii(self):
        """Test saving arrays in a ascii file."""
        array = np.array([0, 1, 3])
        errors = np.array([1, 1, 1])
        save_as_ascii(
            [array, errors],
            filename=os.path.join(self.datadir, "bububu.txt"),
            colnames=["array", "err"],
        )

    def test_save_as_ascii_too_many_dims(self):
        """Test saving arrays in a ascii file."""
        array = np.array([0, 1, 3])
        errors = np.array([1, 1, 1])
        dummy_out = "bububu_bad.txt"
        retval = save_as_ascii(
            np.array([[array], [errors]]),
            filename=dummy_out,
            colnames=["array", "err"],
        )
        assert not os.path.exists(dummy_out)
        assert retval == -1

    def test_save_as_ascii_simple_array(self):
        """Test saving arrays in a ascii file."""
        array = np.array([0, 1, 3])
        dummy_out = "bububu_good.txt"
        retval = save_as_ascii(array, filename=dummy_out, colnames=["array"])
        assert os.path.exists(dummy_out)
        os.unlink(dummy_out)

    @classmethod
    def teardown_class(cls):
        import shutil

        for dum in glob.glob("bubu*.*"):
            os.unlink(dum)
        if os.path.exists("bubu"):
            shutil.rmtree("bubu")


class TestIOModel:
    """Real unit tests."""

    @classmethod
    def setup_class(cls):
        cls.dum = "bubu" + HEN_FILE_EXTENSION
        cls.model = models.Gaussian1D() + models.Const1D(amplitude=2)

    def test_save_and_load_Astropy_model(self):
        save_model(self.model, "bubu_model.p")
        b, kind, _ = load_model("bubu_model.p")
        assert kind == "Astropy"
        assert isinstance(b, Model)
        assert np.all(self.model.parameters == b.parameters)
        assert np.all(self.model.bounds == b.bounds)
        assert np.all(self.model.fixed == b.fixed)

    def test_save_and_load_callable_model(self):
        constraints0 = {"bounds": ()}
        save_model(_dummy, "bubu_callable.p", constraints=constraints0)
        b, kind, constraints = load_model("bubu_callable.p")
        assert kind == "callable"
        assert callable(b)
        assert np.all(_dummy.__code__.co_argcount == b.__code__.co_argcount)
        assert np.all(_dummy.__defaults__ == b.__defaults__)
        assert np.all(constraints == constraints0)

    def test_save_callable_model_wrong(self):
        with pytest.raises(TypeError) as record:
            save_model(_dummy_bad, "callable_bad.p")
        assert "Accepted callable models have only" in str(record.value)
        assert not os.path.exists("callable_bad.p")

    def test_save_junk_model(self):
        a = "g"
        with pytest.raises(TypeError) as record:
            save_model(a, "bad.p", constraints={"bounds": ()})
        assert "The model has to be an Astropy model or a callable" in str(
            record.value
        )
        assert not os.path.exists("bad.p")

    def test_load_python_model_callable(self):
        modelstring = """
def model(x, a=2, b=4):
    return x * a + b

constraints = {'fixed': {'a': True}}
"""
        with open("bubu__model__.py", "w") as fobj:
            print(modelstring, file=fobj)
        b, kind, constraints = load_model("bubu__model__.py")
        assert kind == "callable"
        assert callable(b)
        assert b.__code__.co_argcount == 3
        assert b.__defaults__[0] == 2
        assert b.__defaults__[1] == 4
        assert np.all(constraints == {"fixed": {"a": True}})

    def test_load_python_model_Astropy(self):
        modelstring = """
from astropy.modeling import models
model = models.Const1D()
"""
        with open("bubu__model__2__.py", "w") as fobj:
            print(modelstring, file=fobj)
        b, kind, constraints = load_model("bubu__model__2__.py")
        assert isinstance(b, Model)
        assert kind == "Astropy"
        assert b.amplitude == 1

    def test_load_model_input_not_string(self):
        """Input is not a string"""
        with pytest.raises(TypeError) as record:
            b, kind, _ = load_model(1)
        assert "modelstring has to be an existing file name" in str(
            record.value
        )

    def test_load_model_input_file_doesnt_exist(self):
        with pytest.raises(FileNotFoundError) as record:
            b, kind, _ = load_model("dfasjkdaslfj")
        assert "Model file not found" in str(record.value)

    def test_load_model_input_invalid_file_format(self):
        with open("bubu.txt", "w") as fobj:
            print(1, file=fobj)
        with pytest.raises(TypeError) as record:
            b, kind, _ = load_model("bubu.txt")
        assert "Unknown file type" in str(record.value)

    def test_load_data_fails(self):
        with pytest.raises(TypeError) as record:
            load_data("afile.fits")
        assert "The file type is not recognized" in str(record.value)

    @classmethod
    def teardown_class(cls):
        import shutil

        for dum in glob.glob("bubu*.*"):
            os.unlink(dum)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""A miscellaneous collection of basic functions."""

import sys
import copy
import os
import warnings
from collections.abc import Iterable
from pathlib import Path
import tempfile

import numpy as np
from numpy import histogram2d as histogram2d_np
from numpy import histogram as histogram_np
from astropy.logger import AstropyUserWarning
from astropy import log
from stingray.stats import pds_probability, pds_detection_level
from stingray.stats import z2_n_detection_level, z2_n_probability
from stingray.stats import fold_detection_level, fold_profile_probability
from stingray.pulse.pulsar import _load_and_prepare_TOAs

try:
    import pint.toa as toa
    import pint
    from pint.models import get_model

    HAS_PINT = True
except ImportError:
    HAS_PINT = False

try:
    from skimage.feature import peak_local_max

    HAS_SKIMAGE = True
except ImportError:
    HAS_SKIMAGE = False

try:
    from tqdm import tqdm as show_progress
except ImportError:

    def show_progress(a):
        return a


from . import (
    prange,
    array_take,
    HAS_NUMBA,
    njit,
    vectorize,
    float32,
    float64,
    int32,
    int64,
)

__all__ = [
    "array_take",
    "njit",
    "prange",
    "show_progress",
    "z2_n_detection_level",
    "z2_n_probability",
    "pds_detection_level",
    "pds_probability",
    "fold_detection_level",
    "fold_profile_probability",
    "r_in",
    "r_det",
    "_assign_value_if_none",
    "_look_for_array_in_array",
    "is_string",
    "_order_list_of_arrays",
    "mkdir_p",
    "common_name",
    "hen_root",
    "optimal_bin_time",
    "gti_len",
    "deorbit_events",
    "_add_default_args",
    "check_negative_numbers_in_args",
    "interpret_bintime",
    "get_bin_edges",
    "compute_bin",
    "hist1d_numba_seq",
    "hist2d_numba_seq",
    "hist3d_numba_seq",
    "hist2d_numba_seq_weight",
    "hist3d_numba_seq_weight",
    "index_arr",
    "index_set_arr",
    "histnd_numba_seq",
    "histogram2d",
    "histogram",
    "touch",
    "log_x",
    "get_list_of_small_powers",
    "adjust_dt_for_power_of_two",
    "adjust_dt_for_small_power",
    "memmapped_arange",
    "nchars_in_int_value",
]


DEFAULT_PARSER_ARGS = {}
DEFAULT_PARSER_ARGS["loglevel"] = dict(
    args=["--loglevel"],
    kwargs=dict(
        help=(
            "use given logging level (one between INFO, "
            "WARNING, ERROR, CRITICAL, DEBUG; "
            "default:WARNING)"
        ),
        default="WARNING",
        type=str,
    ),
)
DEFAULT_PARSER_ARGS["nproc"] = dict(
    args=["--nproc"],
    kwargs=dict(help=("Number of processors to use"), default=1, type=int),
)
DEFAULT_PARSER_ARGS["debug"] = dict(
    args=["--debug"],
    kwargs=dict(
        help=("set DEBUG logging level"), default=False, action="store_true"
    ),
)
DEFAULT_PARSER_ARGS["bintime"] = dict(
    args=["-b", "--bintime"],
    kwargs=dict(help="Bin time", type=np.longdouble, default=1),
)
DEFAULT_PARSER_ARGS["energies"] = dict(
    args=["-e", "--energy-interval"],
    kwargs=dict(
        help="Energy interval used for filtering",
        nargs=2,
        type=float,
        default=None,
    ),
)
DEFAULT_PARSER_ARGS["pi"] = dict(
    args=["--pi-interval"],
    kwargs=dict(
        help="PI interval used for filtering",
        nargs=2,
        type=int,
        default=[-1, -1],
    ),
)
DEFAULT_PARSER_ARGS["deorbit"] = dict(
    args=["-p", "--deorbit-par"],
    kwargs=dict(
        help=(
            "Deorbit data with this parameter file (requires PINT installed)"
        ),
        default=None,
        type=str,
    ),
)
DEFAULT_PARSER_ARGS["output"] = dict(
    args=["-o", "--outfile"],
    kwargs=dict(help="Output file", default=None, type=str),
)
DEFAULT_PARSER_ARGS["usepi"] = dict(
    args=["--use-pi"],
    kwargs=dict(
        help="Use the PI channel instead of energies",
        default=False,
        action="store_true",
    ),
)
DEFAULT_PARSER_ARGS["test"] = dict(
    args=["--test"],
    kwargs=dict(
        help="Only used for tests", default=False, action="store_true"
    ),
)
DEFAULT_PARSER_ARGS["pepoch"] = dict(
    args=["--pepoch"],
    kwargs=dict(
        type=float,
        required=False,
        help="Reference epoch for timing parameters (MJD)",
        default=None,
    ),
)


def r_in(td, r_0):
    """Calculate incident countrate given dead time and detected countrate."""
    tau = 1 / r_0
    return 1.0 / (tau - td)


def r_det(td, r_i):
    """Calculate detected countrate given dead time and incident countrate."""
    tau = 1 / r_i
    return 1.0 / (tau + td)


def _assign_value_if_none(value, default):
    if value is None:
        return default
    return value


def _look_for_array_in_array(array1, array2):
    """
    Examples
    --------
    >>> _look_for_array_in_array([1, 2], [2, 3, 4])
    2
    >>> _look_for_array_in_array([1, 2], [3, 4, 5]) is None
    True
    """
    for a1 in array1:
        if a1 in array2:
            return a1
    return None


def is_string(s):
    """Portable function to answer this question."""
    return isinstance(s, str)  # NOQA


def _order_list_of_arrays(data, order):
    """
    Examples
    --------
    >>> order = [1, 2, 0]
    >>> new = _order_list_of_arrays({'a': [4, 5, 6], 'b':[7, 8, 9]}, order)
    >>> np.all(new['a'] == [5, 6, 4])
    True
    >>> np.all(new['b'] == [8, 9, 7])
    True
    >>> new = _order_list_of_arrays([[4, 5, 6], [7, 8, 9]], order)
    >>> np.all(new[0] == [5, 6, 4])
    True
    >>> np.all(new[1] == [8, 9, 7])
    True
    >>> _order_list_of_arrays(2, order) is None
    True
    """
    if hasattr(data, "items"):
        data = dict((i[0], np.asarray(i[1])[order]) for i in data.items())
    elif hasattr(data, "index"):
        data = [np.asarray(i)[order] for i in data]
    else:
        data = None
    return data


class _empty:
    def __init__(self):
        pass


def mkdir_p(path):
    """Safe mkdir function."""
    return os.makedirs(path, exist_ok=True)


def common_name(str1, str2, default="common"):
    """Strip two strings of the letters not in common.

    Filenames must be of same length and only differ by a few letters.

    Parameters
    ----------
    str1 : str
    str2 : str

    Returns
    -------
    common_str : str
        A string containing the parts of the two names in common

    Other Parameters
    ----------------
    default : str
        The string to return if common_str is empty

    Examples
    --------
    >>> common_name('strAfpma', 'strBfpmb')
    'strfpm'
    >>> common_name('strAfpma', 'strBfpmba')
    'common'
    >>> common_name('asdfg', 'qwerr')
    'common'
    >>> common_name('A_3-50_A.nc', 'B_3-50_B.nc')
    '3-50'
    """
    if not len(str1) == len(str2):
        return default
    common_str = ""
    # Extract the HEN root of the name (in case they're event files)
    str1 = hen_root(str1)
    str2 = hen_root(str2)
    for i, letter in enumerate(str1):
        if str2[i] == letter:
            common_str += letter
    # Remove leading and trailing underscores and dashes
    common_str = common_str.rstrip("_").rstrip("-")
    common_str = common_str.lstrip("_").lstrip("-")
    if common_str == "":
        common_str = default
    # log.debug('common_name: %s %s -> %s', str1, str2, common_str)
    return common_str


def hen_root(filename):
    """Return the root file name (without _ev, _lc, etc.).

    Parameters
    ----------
    filename : str
    """
    fname = filename.replace(".gz", "")
    fname = os.path.splitext(fname)[0]
    fname = fname.replace("_ev", "").replace("_lc", "")
    fname = fname.replace("_calib", "")
    return fname


def optimal_bin_time(fftlen, tbin):
    """Vary slightly the bin time to have a power of two number of bins.

    Given an FFT length and a proposed bin time, return a bin time
    slightly shorter than the original, that will produce a power-of-two number
    of FFT bins.

    Examples
    --------
    >>> optimal_bin_time(512, 1.1)
    1.0
    """
    current_nbin = fftlen / tbin
    new_nbin = 2 ** np.ceil(np.log2(current_nbin))
    return fftlen / new_nbin


def gti_len(gti):
    """Return the total good time from a list of GTIs.

    Examples
    --------
    >>> gti_len([[0, 1], [2, 4]])
    3
    """
    return np.sum(np.diff(gti, axis=1))


def simple_orbit_fun_from_parfile(
    mjdstart, mjdstop, parfile, ntimes=1000, ephem="DE421", invert=False
):
    """Get a correction for orbital motion from pulsar parameter file.

    Parameters
    ----------
    mjdstart, mjdstop : float
        Start and end of the time interval where we want the orbital solution
    parfile : str
        Any parameter file understood by PINT (Tempo or Tempo2 format)

    Other parameters
    ----------------
    ntimes : int
        Number of time intervals to use for interpolation. Default 1000
    invert : bool
        Invert the solution (e.g. to apply an orbital model instead of
        subtracting it)

    Returns
    -------
    correction_mjd : function
        Function that accepts times in MJDs and returns the deorbited times.
    """
    from scipy.interpolate import interp1d
    from astropy import units

    if not HAS_PINT:
        raise ImportError(
            "You need the optional dependency PINT to use this "
            "functionality: github.com/nanograv/pint"
        )

    mjds = np.linspace(mjdstart, mjdstop, ntimes)
    toalist = _load_and_prepare_TOAs(mjds, ephem=ephem)
    m = get_model(parfile)
    delays = m.delay(toalist)
    if invert:
        delays = -delays

    correction = interp1d(
        mjds,
        (toalist.table["tdbld"] * units.d - delays).to(units.d).value,
        fill_value="extrapolate",
    )

    return correction


def deorbit_events(events, parameter_file=None, invert=False, ephem=None):
    """Refer arrival times to the center of mass of binary system.

    Parameters
    ----------
    events : `stingray.events.EventList` object
        The event list
    parameter_file : str
        The parameter file, in Tempo-compatible format, containing
        the orbital solution (e.g. a BT model)
    """
    events = copy.deepcopy(events)
    if parameter_file is None:
        warnings.warn(
            "No parameter file specified for deorbit. Returning"
            " unaltered event list"
        )
        return events
    if not os.path.exists(parameter_file):
        raise FileNotFoundError(
            "Parameter file {} does not exist".format(parameter_file)
        )

    if events.mjdref < 33282.0:
        raise ValueError(
            "MJDREF is very low (<01-01-1950), " "this is unsupported."
        )

    pepoch = events.gti[0, 0]
    pepoch_mjd = pepoch / 86400 + events.mjdref

    length = np.max(events.time) - np.min(events.time)
    if length > 200000:
        warnings.warn(
            "The observation is very long. The barycentric correction "
            "will be rough"
        )

    if ephem is None and hasattr(events, "ephem") and events.ephem is not None:
        ephem = events.ephem
        log.info(f"Using default ephemeris: {ephem}")
    elif ephem is None:
        ephem = "DE421"

    length_d = length / 86400
    orbital_correction_fun = simple_orbit_fun_from_parfile(
        pepoch_mjd - 1,
        pepoch_mjd + length_d + 1,
        parameter_file,
        ntimes=min(int(length // 10), 10000),
        invert=invert,
        ephem=ephem,
    )

    mjdtimes = events.time / 86400 + events.mjdref
    mjdgtis = events.gti / 86400 + events.mjdref

    outtime = (orbital_correction_fun(mjdtimes) - events.mjdref) * 86400
    outgtis = (orbital_correction_fun(mjdgtis) - events.mjdref) * 86400

    events.time = outtime
    events.gti = outgtis
    return events


def _add_default_args(parser, list_of_args):
    for key in list_of_args:
        arg = DEFAULT_PARSER_ARGS[key]
        a = arg["args"]
        k = arg["kwargs"]
        parser.add_argument(*a, **k)

    return parser


def check_negative_numbers_in_args(args):
    """If there are negative numbers in args, prepend a space.

    Examples
    --------
    >>> args = ['events.nc', '-f', '103', '--fdot', '-2e-10']
    >>> newargs = check_negative_numbers_in_args(args)
    >>> args[:4] == newargs[:4]
    True
    >>> newargs[4] == ' -2e-10'
    True
    """
    if args is None:
        args = sys.argv[1:]
    newargs = []
    for arg in args:
        try:
            # Has to be a number, has to be negative
            assert -float(arg) > 0
        except (ValueError, AssertionError):
            newargs.append(arg)
            continue

        newargs.append(" " + arg)

    return newargs


def interpret_bintime(bintime):
    """If bin time is negative, interpret as power of two.

    Examples
    --------
    >>> interpret_bintime(2)
    2
    >>> interpret_bintime(-2) == 0.25
    True
    >>> interpret_bintime(0)
    Traceback (most recent call last):
        ...
    ValueError: Bin time cannot be = 0
    """
    if bintime < 0:
        return 2 ** bintime
    elif bintime > 0:
        return bintime
    raise ValueError("Bin time cannot be = 0")


@njit(nogil=True, parallel=False)
def _get_bin_edges(a, bins, a_min, a_max):
    bin_edges = np.zeros(bins + 1, dtype=np.float64)

    delta = (a_max - a_min) / bins
    for i in range(bin_edges.size):
        bin_edges[i] = a_min + i * delta

    bin_edges[-1] = a_max  # Avoid roundoff error on last point
    return bin_edges


def get_bin_edges(a, bins):
    """

    Examples
    --------
    >>> array = np.array([0., 10.])
    >>> bins = 2
    >>> np.allclose(get_bin_edges(array, bins), [0, 5, 10])
    True
    """
    a_min = np.min(a)
    a_max = np.max(a)
    return _get_bin_edges(a, bins, a_min, a_max)


@njit(nogil=True, parallel=False)
def compute_bin(x, bin_edges):
    """

    Examples
    --------
    >>> bin_edges = np.array([0, 5, 10])
    >>> compute_bin(1, bin_edges)
    0
    >>> compute_bin(5, bin_edges)
    1
    >>> compute_bin(10, bin_edges)
    1
    """

    # assuming uniform bins for now
    n = bin_edges.shape[0] - 1
    a_min = bin_edges[0]
    a_max = bin_edges[-1]

    # special case to mirror NumPy behavior for last bin
    if x == a_max:
        return n - 1  # a_max always in last bin

    bin = int(n * (x - a_min) / (a_max - a_min))

    if bin < 0 or bin >= n:
        return None
    else:
        return bin


@njit(nogil=True, parallel=False)
def _hist1d_numba_seq(H, tracks, bins, ranges):
    delta = 1 / ((ranges[1] - ranges[0]) / bins)

    for t in range(tracks.size):
        i = (tracks[t] - ranges[0]) * delta
        if 0 <= i < bins:
            H[int(i)] += 1

    return H


def hist1d_numba_seq(a, bins, ranges, use_memmap=False, tmp=None):
    """
    Examples
    --------
    >>> if os.path.exists('out.npy'): os.unlink('out.npy')
    >>> x = np.random.uniform(0., 1., 100)
    >>> H, xedges = np.histogram(x, bins=5, range=[0., 1.])
    >>> Hn = hist1d_numba_seq(x, bins=5, ranges=[0., 1.], tmp='out.npy',
    ...                       use_memmap=True)
    >>> assert np.all(H == Hn)
    >>> # The number of bins is small, memory map was not used!
    >>> assert not os.path.exists('out.npy')
    >>> H, xedges = np.histogram(x, bins=10**8, range=[0., 1.])
    >>> Hn = hist1d_numba_seq(x, bins=10**8, ranges=[0., 1.], tmp='out.npy',
    ...                       use_memmap=True)
    >>> assert np.all(H == Hn)
    >>> assert os.path.exists('out.npy')
    """
    if bins > 10 ** 7 and use_memmap:
        if tmp is None:
            tmp = tempfile.NamedTemporaryFile("w+")
        hist_arr = np.lib.format.open_memmap(
            tmp, mode="w+", dtype=a.dtype, shape=(bins,)
        )
    else:
        hist_arr = np.zeros((bins,), dtype=a.dtype)

    return _hist1d_numba_seq(hist_arr, a, bins, np.asarray(ranges))


@njit(nogil=True, parallel=False)
def _hist2d_numba_seq(H, tracks, bins, ranges):
    delta = 1 / ((ranges[:, 1] - ranges[:, 0]) / bins)

    for t in range(tracks.shape[1]):
        i = (tracks[0, t] - ranges[0, 0]) * delta[0]
        j = (tracks[1, t] - ranges[1, 0]) * delta[1]
        if 0 <= i < bins[0] and 0 <= j < bins[1]:
            H[int(i), int(j)] += 1

    return H


def hist2d_numba_seq(x, y, bins, ranges):
    """
    Examples
    --------
    >>> x = np.random.uniform(0., 1., 100)
    >>> y = np.random.uniform(2., 3., 100)
    >>> H, xedges, yedges = np.histogram2d(x, y, bins=(5, 5),
    ...                                    range=[(0., 1.), (2., 3.)])
    >>> Hn = hist2d_numba_seq(x, y, bins=(5, 5),
    ...                       ranges=[[0., 1.], [2., 3.]])
    >>> assert np.all(H == Hn)
    """
    H = np.zeros((bins[0], bins[1]), dtype=np.uint64)
    return _hist2d_numba_seq(
        H, np.array([x, y]), np.asarray(list(bins)), np.asarray(ranges)
    )


@njit(nogil=True, parallel=False)
def _hist3d_numba_seq(H, tracks, bins, ranges):
    delta = 1 / ((ranges[:, 1] - ranges[:, 0]) / bins)

    for t in range(tracks.shape[1]):
        i = (tracks[0, t] - ranges[0, 0]) * delta[0]
        j = (tracks[1, t] - ranges[1, 0]) * delta[1]
        k = (tracks[2, t] - ranges[2, 0]) * delta[2]
        if 0 <= i < bins[0] and 0 <= j < bins[1]:
            H[int(i), int(j), int(k)] += 1

    return H


def hist3d_numba_seq(tracks, bins, ranges):
    """
    Examples
    --------
    >>> x = np.random.uniform(0., 1., 100)
    >>> y = np.random.uniform(2., 3., 100)
    >>> z = np.random.uniform(4., 5., 100)
    >>> H, _ = np.histogramdd((x, y, z), bins=(5, 6, 7),
    ...                       range=[(0., 1.), (2., 3.), (4., 5)])
    >>> Hn = hist3d_numba_seq((x, y, z), bins=(5, 6, 7),
    ...                       ranges=[[0., 1.], [2., 3.], [4., 5.]])
    >>> assert np.all(H == Hn)
    """

    H = np.zeros((bins[0], bins[1], bins[2]), dtype=np.uint64)
    return _hist3d_numba_seq(
        H, np.asarray(tracks), np.asarray(list(bins)), np.asarray(ranges)
    )


@njit(nogil=True, parallel=False)
def _hist2d_numba_seq_weight(H, tracks, weights, bins, ranges):
    delta = 1 / ((ranges[:, 1] - ranges[:, 0]) / bins)

    for t in range(tracks.shape[1]):
        i = (tracks[0, t] - ranges[0, 0]) * delta[0]
        j = (tracks[1, t] - ranges[1, 0]) * delta[1]
        if 0 <= i < bins[0] and 0 <= j < bins[1]:
            H[int(i), int(j)] += weights[t]

    return H


def hist2d_numba_seq_weight(x, y, weights, bins, ranges):
    """
    Examples
    --------
    >>> x = np.random.uniform(0., 1., 100)
    >>> y = np.random.uniform(2., 3., 100)
    >>> weight = np.random.uniform(0, 1, 100)
    >>> H, xedges, yedges = np.histogram2d(x, y, bins=(5, 5),
    ...                                    range=[(0., 1.), (2., 3.)],
    ...                                    weights=weight)
    >>> Hn = hist2d_numba_seq_weight(x, y, bins=(5, 5),
    ...                              ranges=[[0., 1.], [2., 3.]],
    ...                              weights=weight)
    >>> assert np.all(H == Hn)
    """
    H = np.zeros((bins[0], bins[1]), dtype=np.double)
    return _hist2d_numba_seq_weight(
        H,
        np.array([x, y]),
        weights,
        np.asarray(list(bins)),
        np.asarray(ranges),
    )


@njit(nogil=True, parallel=False)
def _hist3d_numba_seq_weight(H, tracks, weights, bins, ranges):
    delta = 1 / ((ranges[:, 1] - ranges[:, 0]) / bins)

    for t in range(tracks.shape[1]):
        i = (tracks[0, t] - ranges[0, 0]) * delta[0]
        j = (tracks[1, t] - ranges[1, 0]) * delta[1]
        k = (tracks[2, t] - ranges[2, 0]) * delta[2]
        if 0 <= i < bins[0] and 0 <= j < bins[1]:
            H[int(i), int(j), int(k)] += weights[t]

    return H


def hist3d_numba_seq_weight(tracks, weights, bins, ranges):
    """
    Examples
    --------
    >>> x = np.random.uniform(0., 1., 100)
    >>> y = np.random.uniform(2., 3., 100)
    >>> z = np.random.uniform(4., 5., 100)
    >>> weights = np.random.uniform(0, 1., 100)
    >>> H, _ = np.histogramdd((x, y, z), bins=(5, 6, 7),
    ...                       range=[(0., 1.), (2., 3.), (4., 5)],
    ...                       weights=weights)
    >>> Hn = hist3d_numba_seq_weight(
    ...    (x, y, z), weights, bins=(5, 6, 7),
    ...    ranges=[[0., 1.], [2., 3.], [4., 5.]])
    >>> assert np.all(H == Hn)
    """

    H = np.zeros((bins[0], bins[1], bins[2]), dtype=np.double)
    return _hist3d_numba_seq_weight(
        H,
        np.asarray(tracks),
        weights,
        np.asarray(list(bins)),
        np.asarray(ranges),
    )


@njit(nogil=True, parallel=False)
def index_arr(a, ix_arr):
    strides = np.array(a.strides) / a.itemsize
    ix = int((ix_arr * strides).sum())
    return a.ravel()[ix]


@njit(nogil=True, parallel=False)
def index_set_arr(a, ix_arr, val):
    strides = np.array(a.strides) / a.itemsize
    ix = int((ix_arr * strides).sum())
    a.ravel()[ix] = val


@njit(nogil=True, parallel=False)
def _histnd_numba_seq(H, tracks, bins, ranges, slice_int):
    delta = 1 / ((ranges[:, 1] - ranges[:, 0]) / bins)

    for t in range(tracks.shape[1]):
        slicearr = np.array(
            [
                (tracks[dim, t] - ranges[dim, 0]) * delta[dim]
                for dim in range(tracks.shape[0])
            ]
        )

        good = np.all((slicearr < bins) & (slicearr >= 0))
        slice_int[:] = slicearr

        if good:
            curr = index_arr(H, slice_int)
            index_set_arr(H, slice_int, curr + 1)

    return H


def histnd_numba_seq(tracks, bins, ranges):
    """
    Examples
    --------
    >>> x = np.random.uniform(0., 1., 100)
    >>> y = np.random.uniform(2., 3., 100)
    >>> z = np.random.uniform(4., 5., 100)
    >>> # 2d example
    >>> H, _, _ = np.histogram2d(x, y, bins=np.array((5, 5)),
    ...                          range=[(0., 1.), (2., 3.)])
    >>> alldata = np.array([x, y])
    >>> Hn = histnd_numba_seq(alldata, bins=np.array([5, 5]),
    ...                       ranges=np.array([[0., 1.], [2., 3.]]))
    >>> assert np.all(H == Hn)
    >>> # 3d example
    >>> H, _ = np.histogramdd((x, y, z), bins=np.array((5, 6, 7)),
    ...                       range=[(0., 1.), (2., 3.), (4., 5)])
    >>> alldata = np.array([x, y, z])
    >>> Hn = hist3d_numba_seq(alldata, bins=np.array((5, 6, 7)),
    ...                       ranges=np.array([[0., 1.], [2., 3.], [4., 5.]]))
    >>> assert np.all(H == Hn)
    """
    H = np.zeros(tuple(bins), dtype=np.uint64)
    slice_int = np.zeros(len(bins), dtype=np.uint64)

    return _histnd_numba_seq(H, tracks, bins, ranges, slice_int)


if HAS_NUMBA:

    def histogram2d(*args, **kwargs):
        if "range" in kwargs:
            kwargs["ranges"] = kwargs.pop("range")
        return hist2d_numba_seq(*args, **kwargs)

    def histogram(*args, **kwargs):
        if "range" in kwargs:
            kwargs["ranges"] = kwargs.pop("range")
        return hist1d_numba_seq(*args, **kwargs)


else:

    def histogram2d(*args, **kwargs):
        return histogram2d_np(*args, **kwargs)[0]

    def histogram(*args, **kwargs):
        return histogram_np(*args, **kwargs)[0]


def touch(fname):
    """Mimick the same shell command.

    Examples
    --------
    >>> touch('bububu')
    >>> os.path.exists('bububu')
    True
    >>> os.unlink('bububu')
    """
    Path(fname).touch()


def log_x(a, base):
    # Logb x = Loga x/Loga b
    return np.log(a) / np.log(base)


def get_list_of_small_powers(maxno=100000000000):
    powers_of_two = 2 ** np.arange(0, np.ceil(np.log2(maxno)))
    powers_of_three = 3 ** np.arange(0, np.ceil(log_x(maxno, 3)))
    powers_of_five = 5 ** np.arange(0, np.ceil(log_x(maxno, 5)))
    list_of_powers = []
    for p2 in powers_of_two:
        for p3 in powers_of_three:
            for p5 in powers_of_five:
                newno = p2 * p3 * p5
                if newno > maxno:
                    break
                list_of_powers.append(p2 * p3 * p5)
    return sorted(list_of_powers)


def adjust_dt_for_power_of_two(dt, length, strict=False):
    """
    Examples
    --------
    >>> length, dt = 10, 0.1
    >>> # There are 100 bins there. I want them to be 128.
    >>> new_dt = adjust_dt_for_power_of_two(dt, length)
    INFO: ...
    INFO: ...
    >>> np.isclose(new_dt, 0.078125)
    True
    >>> length / new_dt == 128
    True
    >>> length, dt = 6.5, 0.1
    >>> # There are 100 bins there. I want them to be 128.
    >>> new_dt = adjust_dt_for_power_of_two(dt, length)
    INFO: ...
    INFO: Too many ...
    INFO: ...
    >>> length / new_dt == 72
    True
    """
    log.info("Adjusting bin time to closest power of 2 of bins.")
    nbin = length / dt
    closest_to_pow2 = 2 ** np.ceil(np.log2(nbin))
    if closest_to_pow2 > 1.5 * nbin and not strict:
        log.info("Too many bins: using powers of 2, 3, and 5.")
        return adjust_dt_for_small_power(dt, length)
    new_dt = length / closest_to_pow2
    log.info(f"New bin time: {new_dt} (nbin {nbin} -> {closest_to_pow2})")
    return new_dt


def adjust_dt_for_small_power(dt, length):
    """
    Examples
    --------
    >>> length, dt = 9.9, 0.1
    >>> # There are 99 bins there. I want them to be 100 (2**2 * 5**2).
    >>> new_dt = adjust_dt_for_small_power(dt, length)
    INFO:...
    >>> np.isclose(new_dt, 0.099)
    True
    >>> np.isclose(length / new_dt, 100)
    True
    """
    nbin = length / dt
    losp = get_list_of_small_powers(2 * nbin)
    nbin_new = np.searchsorted(losp, nbin)
    if losp[nbin_new] < nbin:
        nbin_new += 1

    new_dt = length / losp[nbin_new]
    log.info(f"New bin time: {new_dt} (nbin {nbin} -> {losp[nbin_new]})")
    return new_dt


def memmapped_arange(
    i0, i1, istep, fname=None, nbin_threshold=10 ** 7, dtype=float
):
    """Arange plus memory mapping.

    Examples
    --------
    >>> i0, i1, istep = 0, 10, 1e-2
    >>> np.allclose(np.arange(i0, i1, istep), memmapped_arange(i0, i1, istep))
    True
    >>> i0, i1, istep = 0, 10, 1e-7
    >>> np.allclose(np.arange(i0, i1, istep), memmapped_arange(i0, i1, istep))
    True

    """
    import tempfile

    chunklen = 10 ** 6
    Nbins = int((i1 - i0) / istep)
    if Nbins < nbin_threshold:
        return np.arange(i0, i1, istep)
    if fname is None:
        _, fname = tempfile.mkstemp(suffix=".npy")

    hist_arr = np.lib.format.open_memmap(
        fname, mode="w+", dtype=dtype, shape=(Nbins,)
    )

    for start in range(0, Nbins, chunklen):
        stop = min(start + chunklen, Nbins)
        hist_arr[start:stop] = np.arange(start, stop) * istep

    return hist_arr


def nchars_in_int_value(value):
    """Number of characters to write an integer number

    Examples
    --------
    >>> nchars_in_int_value(2)
    1
    >>> nchars_in_int_value(1356)
    4
    >>> nchars_in_int_value(9999)
    4
    >>> nchars_in_int_value(10000)
    5
    """
    #  "+1" because, e.g., 10000 would return 4
    return int(np.ceil(np.log10(value + 1)))


def find_peaks_in_image(image, n=5, rough=False, **kwargs):
    """

    Parameters
    ----------
    image : :class:`np.array`
        An image

    Other Parameters
    ----------------
    n: int
        Number of best peaks to find
    kwargs: dict
        Additional parameters to be passed to skimage

    Returns
    -------
    idxs : list of tuples
        List of indices of the maxima

    Examples
    --------
    >>> image = np.random.normal(0, 0.01, (10, 10))
    >>> image[5, 5] = 2
    >>> image[7, 2] = 1
    >>> image[8, 1] = 3
    >>> idxs = find_peaks_in_image(image, n=2)
    >>> np.allclose(idxs, [(8, 1), (5, 5)])
    True
    >>> idxs = find_peaks_in_image(image, n=2, rough=True)
    >>> np.allclose(idxs, [(8, 1), (5, 5)])
    True
    """
    if not HAS_SKIMAGE or rough:
        best_cands = [
            np.unravel_index(idx, image.shape)
            for idx in np.argpartition(image.flatten(), -n)[-n:]
        ]
        __best_stats = [image[bci] for bci in best_cands]
        best_cands = np.asarray(best_cands)[np.argsort(__best_stats)][::-1]
        return best_cands

    return peak_local_max(image, num_peaks=n, **kwargs)


def force_iterable(val):
    """Force a number to become an array with one element.

    Arrays are preserved.

    Examples
    --------
    >>> val = 5.
    >>> force_iterable(val)[0] == val
    True
    >>> val = None
    >>> force_iterable(val) is None
    True
    >>> val = np.array([5., 5])
    >>> np.all(force_iterable(val) == val)
    True
    """
    if val is None:
        return val

    if not isinstance(val, Iterable):
        return np.array([val])

    return np.asarray(val)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Functions to create and apply GTIs."""

import sys
import warnings
import numpy as np
from astropy import log
from astropy.logger import AstropyUserWarning
from stingray.gti import cross_gtis, create_gti_from_condition, create_gti_mask
from .io import HEN_FILE_EXTENSION, save_data, load_data, get_file_type
from .base import hen_root, _assign_value_if_none


def filter_gti_by_length(gti, minimum_length):
    """Filter a list of GTIs: keep those longer than `minimum_length`."""
    if minimum_length == 0 or minimum_length is None:
        return gti

    newgtis = []
    for g in gti:
        length = g[1] - g[0]
        if length >= minimum_length:
            newgtis.append(g)

    return np.array(newgtis)


def create_gti(
    fname, filter_expr, safe_interval=[0, 0], outfile=None, minimum_length=0
):
    """Create a GTI list by using boolean operations on file data.

    Parameters
    ----------
    fname : str
        File name. The file must be in HENDRICS format.
    filter_expr : str
        A boolean condition on one or more of the arrays contained in the data.
        E.g. '(lc > 10) & (lc < 20)'

    Returns
    -------
    gtis : [[gti0_0, gti0_1], [gti1_0, gti1_1], ...]
        The newly created GTIs

    Other parameters
    ----------------
    safe_interval : float or [float, float]
        A safe interval to exclude at both ends (if single float) or the start
        and the end (if pair of values) of GTIs.
    outfile : str
        The output file name. If None, use a default root + '_gti' combination
    """
    # Necessary as nc variables are sometimes defined as array
    from numpy import array  # NOQA

    ftype, data = get_file_type(fname, raw_data=True)

    instr = data["instr"]
    if ftype == "lc" and instr.lower() == "pca":
        warnings.warn(
            "RXTE/PCA data; normalizing lc per no. PCUs", AstropyUserWarning
        )
        # If RXTE, plot per PCU count rate
        data["counts"] /= data["nPCUs"]
    mjdref = data["mjdref"]
    # Map all entries of data to local variables
    locals().update(data)

    good = eval(filter_expr)

    gtis = create_gti_from_condition(
        locals()["time"], good, safe_interval=safe_interval
    )

    gtis = filter_gti_by_length(gtis, minimum_length)

    outfile = _assign_value_if_none(
        outfile, hen_root(fname) + "_gti" + HEN_FILE_EXTENSION
    )
    save_data(
        {"gti": gtis, "mjdref": mjdref, "__sr__class__type__": "gti"}, outfile
    )

    return gtis


def apply_gti(fname, gti, outname=None, minimum_length=0):
    """Apply a GTI list to the data contained in a file.

    File MUST have a GTI extension already, and an extension called `time`.
    """
    ftype, data = get_file_type(fname, raw_data=True)

    try:
        datagti = data["gti"]
        newgtis = cross_gtis([gti, datagti])
    except Exception:  # pragma: no cover
        warnings.warn("Data have no GTI extension", AstropyUserWarning)
        newgtis = gti

    newgtis = filter_gti_by_length(newgtis, minimum_length)

    data["gti"] = newgtis
    good = create_gti_mask(data["time"], newgtis)

    data["time"] = data["time"][good]
    if ftype == "lc":
        data["counts"] = data["counts"][good]
        data["counts_err"] = data["counts_err"][good]
    elif ftype == "events":
        for key in ["pi", "energy"]:
            if key in data:
                data[key] = data[key][good]

        if data["instr"] == "PCA":  # pragma: no cover
            data["PCU"] = data["PCU"][good]

    newext = "_gtifilt" + HEN_FILE_EXTENSION
    outname = _assign_value_if_none(
        outname, fname.replace(HEN_FILE_EXTENSION, "") + newext
    )
    save_data(data, outname)

    return newgtis


def main(args=None):
    """Main function called by the `HENcreategti` command line script."""
    import argparse
    from .base import _add_default_args, check_negative_numbers_in_args

    description = (
        "Create GTI files from a filter expression, or applies "
        "previously created GTIs to a file"
    )
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument("files", help="List of files", nargs="+")

    parser.add_argument(
        "-f",
        "--filter",
        type=str,
        default=None,
        help="Filter expression, that has to be a valid "
        + "Python boolean operation on a data variable "
        + "contained in the files",
    )

    parser.add_argument(
        "-c",
        "--create-only",
        default=False,
        action="store_true",
        help="If specified, creates GTIs withouth applying"
        + "them to files (Default: False)",
    )

    parser.add_argument(
        "--overwrite",
        default=False,
        action="store_true",
        help="Overwrite original file (Default: False)",
    )

    parser.add_argument(
        "-a",
        "--apply-gti",
        type=str,
        default=None,
        help="Apply a GTI from this file to input files",
    )

    parser.add_argument(
        "-l",
        "--minimum-length",
        type=float,
        default=0,
        help=(
            "Minimum length of GTIs (below this length, they"
            " will be discarded)"
        ),
    )

    parser.add_argument(
        "--safe-interval",
        nargs=2,
        type=float,
        default=[0, 0],
        help="Interval at start and stop of GTIs used" + " for filtering",
    )

    args = check_negative_numbers_in_args(args)
    _add_default_args(parser, ["loglevel", "debug"])

    args = parser.parse_args(args)
    files = args.files

    if args.debug:
        args.loglevel = "DEBUG"

    log.setLevel(args.loglevel)
    with log.log_to_file("HENcreategti.log"):
        filter_expr = args.filter
        if filter_expr is None and args.apply_gti is None:
            sys.exit(
                "Please specify filter expression (-f option) or input "
                "GTI file (-a option)"
            )

        for fname in files:
            if args.apply_gti is not None:
                data = load_data(args.apply_gti)
                gtis = data["gti"]
            else:
                gtis = create_gti(
                    fname,
                    filter_expr,
                    safe_interval=args.safe_interval,
                    minimum_length=args.minimum_length,
                )
            if args.create_only:
                continue
            if args.overwrite:
                outname = fname
            else:
                # Use default
                outname = None
            apply_gti(
                fname,
                gtis,
                outname=outname,
                minimum_length=args.minimum_length,
            )
